# Copyright (c) 2022 NVIDIA CORPORATION.  All rights reserved.
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an express
# license agreement from NVIDIA CORPORATION is strictly prohibited.

# include parent path
import os
import sys
import numpy as np
import math
import ctypes

import warp as wp
from warp.tests.test_base import *

import unittest

wp.init()


@wp.kernel
def test_conditional_if_else():

    a = 0.5
    b = 2.0

    if a > b:
        c = 1.0
    else:
        c = -1.0

    wp.expect_eq(c, -1.0)
    

@wp.kernel
def test_conditional_if_else_nested():

    a = 1.0
    b = 2.0

    if a > b:

        c = 3.0
        d = 4.0

        if (c > d):
            e = 1.0
        else:
            e = -1.0

    else:

        c = 6.0
        d = 7.0

        if (c > d):
            e = 2.0
        else:
            e = -2.0

    wp.expect_eq(e, -2.0)

@wp.kernel
def test_boolean_and():

    a = 1.0
    b = 2.0
    c = 1.0

    if a > 0.0 and b > 0.0:
        c = -1.0

    wp.expect_eq(c, -1.0)

@wp.kernel
def test_boolean_or():

    a = 1.0
    b = 2.0
    c = 1.0

    if a > 0.0 and b > 0.0:
        c = -1.0

    wp.expect_eq(c, -1.0)
    

@wp.kernel
def test_boolean_compound():

    a = 1.0
    b = 2.0
    c = 3.0
    
    d = 1.0

    if a > 0.0 and b > 0.0 or c > a:
        d = -1.0

    wp.expect_eq(d, -1.0)

@wp.kernel
def test_boolean_literal():

    t = True
    f = False
    
    r = 1.0

    if t == (not f):
        r = -1.0

    wp.expect_eq(r, -1.0)


def register(parent):

    devices = wp.get_devices()

    class TestConditional(parent):
        pass

    add_kernel_test(TestConditional, kernel=test_conditional_if_else, dim=1, devices=devices)
    add_kernel_test(TestConditional, kernel=test_conditional_if_else_nested, dim=1, devices=devices)
    add_kernel_test(TestConditional, kernel=test_boolean_and, dim=1, devices=devices)
    add_kernel_test(TestConditional, kernel=test_boolean_or, dim=1, devices=devices)
    add_kernel_test(TestConditional, kernel=test_boolean_compound, dim=1, devices=devices)
    add_kernel_test(TestConditional, kernel=test_boolean_literal, dim=1, devices=devices)
    add_kernel_test(TestConditional, kernel=test_boolean_and, dim=1, devices=devices)

    return TestConditional


if __name__ == '__main__':
    c = register(unittest.TestCase)
    unittest.main(verbosity=2)



# Copyright (c) 2022 NVIDIA CORPORATION.  All rights reserved.
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an express
# license agreement from NVIDIA CORPORATION is strictly prohibited.

import warp as wp


# decompose a quaternion into a sequence of 3 rotations around x,y',z' respectively, i.e.: q = q_z''q_y'q_x
@wp.func
def quat_decompose(q: wp.quat):

    R = wp.mat33(
            wp.quat_rotate(q, wp.vec3(1.0, 0.0, 0.0)),
            wp.quat_rotate(q, wp.vec3(0.0, 1.0, 0.0)),
            wp.quat_rotate(q, wp.vec3(0.0, 0.0, 1.0)))

    # https://www.sedris.org/wg8home/Documents/WG80485.pdf
    phi = wp.atan2(R[1, 2], R[2, 2])
    theta = wp.asin(-R[0, 2])
    psi = wp.atan2(R[0, 1], R[0, 0])

    return -wp.vec3(phi, theta, psi)

@wp.kernel
def eval_articulation_fk(
    articulation_start: wp.array(dtype=int),
    articulation_mask: wp.array(dtype=int), # used to enable / disable FK for an articulation, if None then treat all as enabled
    joint_q: wp.array(dtype=float),
    joint_qd: wp.array(dtype=float),
    joint_q_start: wp.array(dtype=int),
    joint_qd_start: wp.array(dtype=int),
    joint_type: wp.array(dtype=int),
    joint_parent: wp.array(dtype=int),
    joint_X_p: wp.array(dtype=wp.transform),
    joint_X_c: wp.array(dtype=wp.transform),
    joint_axis: wp.array(dtype=wp.vec3),
    body_com: wp.array(dtype=wp.vec3),
    # outputs
    body_q: wp.array(dtype=wp.transform),
    body_qd: wp.array(dtype=wp.spatial_vector)):

    tid = wp.tid()

    # early out if disabling FK for this articulation
    if (articulation_mask):
        if (articulation_mask[tid]==0):
            return

    joint_start = articulation_start[tid]
    joint_end = articulation_start[tid+1]

    for i in range(joint_start, joint_end):

        parent = joint_parent[i]
        X_wp = wp.transform_identity()
        v_wp = wp.spatial_vector()

        if (parent >= 0):
            X_wp = body_q[parent]
            v_wp = body_qd[parent]

        # compute transform across the joint
        type = joint_type[i]
        axis = joint_axis[i]

        X_pj = joint_X_p[i]
        X_cj = joint_X_c[i]  
        
        q_start = joint_q_start[i]
        qd_start = joint_qd_start[i]

        if type == wp.sim.JOINT_PRISMATIC:

            q = joint_q[q_start]
            qd = joint_qd[qd_start]

            X_jc = wp.transform(axis*q, wp.quat_identity())
            v_jc = wp.spatial_vector(wp.vec3(), axis*qd)

        if type == wp.sim.JOINT_REVOLUTE:

            q = joint_q[q_start]
            qd = joint_qd[qd_start]

            X_jc = wp.transform(wp.vec3(), wp.quat_from_axis_angle(axis, q))
            v_jc = wp.spatial_vector(axis*qd, wp.vec3())

        if type == wp.sim.JOINT_BALL:

            r = wp.quat(joint_q[q_start+0],
                        joint_q[q_start+1],
                        joint_q[q_start+2],
                        joint_q[q_start+3])

            w = wp.vec3(joint_qd[qd_start+0],
                        joint_qd[qd_start+1],
                        joint_qd[qd_start+2])

            # print(r)
            
            X_jc = wp.transform(wp.vec3(), r)
            v_jc = wp.spatial_vector(w, wp.vec3())

        if type == wp.sim.JOINT_FIXED:
            
            X_jc = wp.transform_identity()
            v_jc = wp.spatial_vector(wp.vec3(), wp.vec3())

        if type == wp.sim.JOINT_FREE:

            t = wp.transform(
                    wp.vec3(joint_q[q_start+0], joint_q[q_start+1], joint_q[q_start+2]),
                    wp.quat(joint_q[q_start+3], joint_q[q_start+4], joint_q[q_start+5], joint_q[q_start+6]))

            v = wp.spatial_vector(
                    wp.vec3(joint_qd[qd_start+0], joint_qd[qd_start+1], joint_qd[qd_start+2]),
                    wp.vec3(joint_qd[qd_start+3], joint_qd[qd_start+4], joint_qd[qd_start+5]))

            X_jc = t
            v_jc = v

        if type == wp.sim.JOINT_COMPOUND:

            q_c = wp.transform_get_rotation(X_cj)

            # body local axes
            local_0 = wp.quat_rotate(q_c, wp.vec3(1.0, 0.0, 0.0))
            local_1 = wp.quat_rotate(q_c, wp.vec3(0.0, 1.0, 0.0))
            local_2 = wp.quat_rotate(q_c, wp.vec3(0.0, 0.0, 1.0))

            # reconstruct rotation axes, todo: can probably use fact that rz'*ry'*rx' == rx*ry*rz to avoid some work here
            axis_0 = local_0
            q_0 = wp.quat_from_axis_angle(axis_0, joint_q[q_start+0])

            axis_1 = wp.quat_rotate(q_0, local_1)
            q_1 = wp.quat_from_axis_angle(axis_1, joint_q[q_start+1])

            axis_2 = wp.quat_rotate(q_1*q_0, local_2)
            q_2 = wp.quat_from_axis_angle(axis_2, joint_q[q_start+2])

            t = wp.transform(wp.vec3(), q_2*q_1*q_0)

            v = wp.spatial_vector(axis_0*joint_qd[qd_start+0] + 
                                  axis_1*joint_qd[qd_start+1] + 
                                  axis_2*joint_qd[qd_start+2], wp.vec3())

            X_jc = t
            v_jc = v

        if type == wp.sim.JOINT_UNIVERSAL:

            q_c = wp.transform_get_rotation(X_cj)

            # body local axes
            local_0 = wp.quat_rotate(q_c, wp.vec3(1.0, 0.0, 0.0))
            local_1 = wp.quat_rotate(q_c, wp.vec3(0.0, 1.0, 0.0))

            # reconstruct rotation axes
            axis_0 = local_0
            q_0 = wp.quat_from_axis_angle(axis_0, joint_q[q_start+0])

            axis_1 = wp.quat_rotate(q_0, local_1)
            q_1 = wp.quat_from_axis_angle(axis_1, joint_q[q_start+1])

            t = wp.transform(wp.vec3(), q_1*q_0)

            v = wp.spatial_vector(axis_0*joint_qd[qd_start+0] + 
                                  axis_1*joint_qd[qd_start+1], wp.vec3())

            X_jc = t
            v_jc = v


        X_wj = X_wp*X_pj
        X_wc = X_wj*X_jc

        # transform velocity across the joint to world space
        angular_vel = wp.transform_vector(X_wj, wp.spatial_top(v_jc))
        linear_vel = wp.transform_vector(X_wj, wp.spatial_bottom(v_jc))

        v_wc = v_wp + wp.spatial_vector(angular_vel, linear_vel + wp.cross(angular_vel, body_com[i]))

        body_q[i] = X_wc
        body_qd[i] = v_wc
        

# updates state body information based on joint coordinates
def eval_fk(model, joint_q, joint_qd, mask, state):

    wp.launch(kernel=eval_articulation_fk,
                dim=model.articulation_count,
                inputs=[    
                model.articulation_start,
                mask,
                joint_q,
                joint_qd,
                model.joint_q_start,
                model.joint_qd_start,
                model.joint_type,
                model.joint_parent,
                model.joint_X_p,
                model.joint_X_c,
                model.joint_axis,
                model.body_com],
                outputs=[
                    state.body_q,
                    state.body_qd,
                ],
                device=model.device)


# returns the twist around an axis
@wp.func
def quat_twist(axis: wp.vec3, q: wp.quat):
    
    # project imaginary part onto axis
    a = wp.vec3(q[0], q[1], q[2])
    a = wp.dot(a, axis)*axis

    return wp.normalize(wp.quat(a[0], a[1], a[2], q[3]))


@wp.kernel
def eval_articulation_ik(body_q: wp.array(dtype=wp.transform),
                         body_qd: wp.array(dtype=wp.spatial_vector),
                         body_com: wp.array(dtype=wp.vec3),
                         joint_type: wp.array(dtype=int),
                         joint_parent: wp.array(dtype=int),
                         joint_X_p: wp.array(dtype=wp.transform),
                         joint_X_c: wp.array(dtype=wp.transform),
                         joint_axis: wp.array(dtype=wp.vec3),
                         joint_q_start: wp.array(dtype=int),
                         joint_qd_start: wp.array(dtype=int),
                         joint_q: wp.array(dtype=float),
                         joint_qd: wp.array(dtype=float)):

    tid = wp.tid()

    c_child = tid
    c_parent = joint_parent[tid]

    X_pj = joint_X_p[tid]
    X_cj = joint_X_c[tid]

    X_wp = X_pj
    r_p = wp.vec3()
    w_p = wp.vec3()
    v_p = wp.vec3()

    # parent transform and moment arm
    if (c_parent >= 0):
        X_wp = body_q[c_parent]*X_wp
        r_wp = wp.transform_get_translation(X_wp) - wp.transform_point(body_q[c_parent], body_com[c_parent])
        
        twist_p = body_qd[c_parent]

        w_p = wp.spatial_top(twist_p)
        v_p = wp.spatial_bottom(twist_p) + wp.cross(w_p, r_wp)

    # child transform and moment arm
    X_wc = body_q[c_child]#*joint_X_c[tid]
    r_c = wp.transform_get_translation(X_wc) - wp.transform_point(body_q[c_child], body_com[c_child])
    
    twist_c = body_qd[c_child]

    w_c = wp.spatial_top(twist_c)
    v_c = wp.spatial_bottom(twist_c) + wp.cross(w_c, r_c)

    # joint properties
    type = joint_type[tid]
    axis = joint_axis[tid]
  
    x_p = wp.transform_get_translation(X_wp)
    x_c = wp.transform_get_translation(X_wc)

    q_p = wp.transform_get_rotation(X_wp)
    q_c = wp.transform_get_rotation(X_wc)

    # translational error
    x_err = x_c - x_p
    v_err = v_c - v_p
    w_err = w_c - w_p

    q_start = joint_q_start[tid]
    qd_start = joint_qd_start[tid]

    if type == wp.sim.JOINT_PRISMATIC:
        
        # world space joint axis
        axis_p = wp.transform_vector(X_wp, axis)

        # evaluate joint coordinates
        q = wp.dot(x_err, axis_p)
        qd = wp.dot(v_err, axis_p)

        joint_q[q_start] = q
        joint_qd[qd_start] = qd
        
        return
   
    if type == wp.sim.JOINT_REVOLUTE:
        
        axis_p = wp.transform_vector(X_wp, axis)
        axis_c = wp.transform_vector(X_wc, axis)

        # swing twist decomposition
        q_pc = wp.quat_inverse(q_p)*q_c
        twist = quat_twist(axis, q_pc)

        q = wp.acos(twist[3])*2.0*wp.sign(wp.dot(axis, wp.vec3(twist[0], twist[1], twist[2])))
        qd = wp.dot(w_err, axis_p)

        joint_q[q_start] = q
        joint_qd[qd_start] = qd

        return
    
    if type == wp.sim.JOINT_BALL:
        
        q_pc = wp.quat_inverse(q_p)*q_c
    
        joint_q[q_start + 0] = q_pc[0]
        joint_q[q_start + 1] = q_pc[1]
        joint_q[q_start + 2] = q_pc[2]
        joint_q[q_start + 3] = q_pc[3]

        joint_qd[qd_start + 0] = w_err[0]
        joint_qd[qd_start + 1] = w_err[1]
        joint_qd[qd_start + 2] = w_err[2]

        return

    if type == wp.sim.JOINT_FIXED:
        return

    if type == wp.sim.JOINT_FREE:
        
        q_pc = wp.quat_inverse(q_p)*q_c

        joint_q[q_start + 0] = x_err[0]
        joint_q[q_start + 1] = x_err[1]
        joint_q[q_start + 2] = x_err[2]

        joint_q[q_start + 3] = q_pc[0]
        joint_q[q_start + 4] = q_pc[1]
        joint_q[q_start + 5] = q_pc[2]
        joint_q[q_start + 6] = q_pc[3]
        
        joint_qd[qd_start + 0] = w_err[0]
        joint_qd[qd_start + 1] = w_err[1]
        joint_qd[qd_start + 2] = w_err[2]

        joint_qd[qd_start + 3] = v_err[0]
        joint_qd[qd_start + 4] = v_err[1]
        joint_qd[qd_start + 5] = v_err[2]

    if type == wp.sim.JOINT_COMPOUND:

        q_off = wp.transform_get_rotation(X_cj)
        q_pc = wp.quat_inverse(q_off)*wp.quat_inverse(q_p)*q_c*q_off

        # decompose to a compound rotation each axis 
        angles = quat_decompose(q_pc)

        # reconstruct rotation axes
        axis_0 = wp.vec3(1.0, 0.0, 0.0)
        q_0 = wp.quat_from_axis_angle(axis_0, angles[0])

        axis_1 = wp.quat_rotate(q_0, wp.vec3(0.0, 1.0, 0.0))
        q_1 = wp.quat_from_axis_angle(axis_1, angles[1])

        axis_2 = wp.quat_rotate(q_1*q_0, wp.vec3(0.0, 0.0, 1.0))
        q_2 = wp.quat_from_axis_angle(axis_2, angles[2])

        q_w = q_p*q_off

        joint_q[q_start+0] = angles[0]
        joint_q[q_start+1] = angles[1]
        joint_q[q_start+2] = angles[2]

        joint_qd[qd_start+0] = wp.dot(wp.quat_rotate(q_w, axis_0), w_err)
        joint_qd[qd_start+1] = wp.dot(wp.quat_rotate(q_w, axis_1), w_err)
        joint_qd[qd_start+2] = wp.dot(wp.quat_rotate(q_w, axis_2), w_err)

        return

    if type == wp.sim.JOINT_UNIVERSAL:

        q_off = wp.transform_get_rotation(X_cj)
        q_pc = wp.quat_inverse(q_off)*wp.quat_inverse(q_p)*q_c*q_off

        # decompose to a compound rotation each axis 
        angles = quat_decompose(q_pc)

        # reconstruct rotation axes
        axis_0 = wp.vec3(1.0, 0.0, 0.0)
        q_0 = wp.quat_from_axis_angle(axis_0, angles[0])

        axis_1 = wp.quat_rotate(q_0, wp.vec3(0.0, 1.0, 0.0))
        q_1 = wp.quat_from_axis_angle(axis_1, angles[1])

        q_w = q_p*q_off

        joint_q[q_start+0] = angles[0]
        joint_q[q_start+1] = angles[1]

        joint_qd[qd_start+0] = wp.dot(wp.quat_rotate(q_w, axis_0), w_err)
        joint_qd[qd_start+1] = wp.dot(wp.quat_rotate(q_w, axis_1), w_err)

        return



# given maximal coordinate model computes ik (closest point projection)
def eval_ik(model, state, joint_q, joint_qd):

    wp.launch(kernel=eval_articulation_ik,
        dim=model.body_count,
        inputs=[    
            state.body_q,
            state.body_qd,
            model.body_com,
            model.joint_type,
            model.joint_parent,
            model.joint_X_p,
            model.joint_X_c,
            model.joint_axis,
            model.joint_q_start,
            model.joint_qd_start],
        outputs=[
            joint_q,
            joint_qd
        ],
        device=model.device)




# Copyright (c) 2022 NVIDIA CORPORATION.  All rights reserved.
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an express
# license agreement from NVIDIA CORPORATION is strictly prohibited.

import numpy as np
import warp as wp
from warp.tests.test_base import *

import unittest

wp.init()

@wp.struct
class Model:
    dt: float
    gravity: wp.vec3
    m: wp.array(dtype=float)

@wp.struct
class State:
    x: wp.array(dtype=wp.vec3)
    v: wp.array(dtype=wp.vec3)

@wp.kernel
def kernel_step(state_in: State, state_out: State, model: Model):
    i = wp.tid()

    state_out.v[i] = state_in.v[i] + model.gravity / model.m[i] * model.dt
    state_out.x[i] = state_in.x[i] + state_out.v[i] * model.dt

def test_step(test, device):
    dim = 5

    dt = 0.01
    gravity = np.array([0, 0, -9.81])

    m = np.ones(dim)

    m_model = wp.array(m, dtype=float, device=device)

    model = Model()
    model.m = m_model
    model.dt = dt
    model.gravity = wp.vec3(0, 0, -9.81)

    np.random.seed(0)
    x = np.random.normal(size=(dim, 3))
    v = np.random.normal(size=(dim, 3))

    x_expected = x + (v + gravity / m[:, None] * dt) * dt

    x_in = wp.array(x, dtype=wp.vec3, device=device)
    v_in = wp.array(v, dtype=wp.vec3, device=device)

    state_in = State()
    state_in.x = x_in
    state_in.v = v_in

    state_out = State()
    state_out.x = wp.empty_like(x_in)
    state_out.v = wp.empty_like(v_in)

    with CheckOutput(test):
        wp.launch(kernel_step, dim=dim, inputs=[state_in, state_out, model], device=device)

    assert_np_equal(state_out.x.numpy(), x_expected, tol=1e-6)


@wp.kernel
def kernel_loss(x: wp.array(dtype=wp.vec3), loss: wp.array(dtype=float)):
    i = wp.tid()
    wp.atomic_add(loss, 0, x[i][0] * x[i][0] + x[i][1] * x[i][1] + x[i][2] * x[i][2])


def test_step_grad(test, device):
    dim = 5

    dt = 0.01
    gravity = np.array([0, 0, -9.81])

    np.random.seed(0)
    m = np.random.rand(dim) + 0.1

    m_model = wp.array(m, dtype=float, device=device, requires_grad=True)

    model = Model()
    model.m = m_model
    model.dt = dt
    model.gravity = wp.vec3(0, 0, -9.81)

    x = np.random.normal(size=(dim, 3))
    v = np.random.normal(size=(dim, 3))

    x_in = wp.array(x, dtype=wp.vec3, device=device, requires_grad=True)
    v_in = wp.array(v, dtype=wp.vec3, device=device, requires_grad=True)

    state_in = State()
    state_in.x = x_in
    state_in.v = v_in

    state_out = State()
    state_out.x = wp.empty_like(x_in, requires_grad=True)
    state_out.v = wp.empty_like(v_in, requires_grad=True)

    loss = wp.empty(1, dtype=float, device=device, requires_grad=True)

    tape = wp.Tape()

    with tape:
        wp.launch(kernel_step, dim=dim, inputs=[state_in, state_out, model], device=device)
        wp.launch(kernel_loss, dim=dim, inputs=[state_out.x, loss], device=device)

    tape.backward(loss)

    dl_dx = 2 * state_out.x.numpy()
    dl_dv = dl_dx * dt

    dv_dm = - gravity * dt / m[:, None] ** 2
    dl_dm = (dl_dv * dv_dm).sum(-1)

    assert_np_equal(state_out.x.grad.numpy(), dl_dx, tol=1e-6)
    assert_np_equal(state_in.x.grad.numpy(), dl_dx, tol=1e-6)
    assert_np_equal(state_out.v.grad.numpy(), dl_dv, tol=1e-6)
    assert_np_equal(state_in.v.grad.numpy(), dl_dv, tol=1e-6)
    assert_np_equal(model.m.grad.numpy(), dl_dm, tol=1e-6)

    tape.zero()
    
    assert state_out.x.grad.numpy().sum() == 0.0
    assert state_in.x.grad.numpy().sum() == 0.0
    assert state_out.v.grad.numpy().sum() == 0.0
    assert state_in.v.grad.numpy().sum() == 0.0
    assert model.m.grad.numpy().sum() == 0.0


def register(parent):
    devices = wp.get_devices()
    class TestStruct(parent):
        pass

    add_function_test(TestStruct, "test_step", test_step, devices=devices)
    add_function_test(TestStruct, "test_step_grad", test_step_grad, devices=devices)

    return TestStruct

if __name__ == '__main__':
    c = register(unittest.TestCase)
    unittest.main(verbosity=2)



# Copyright (c) 2022 NVIDIA CORPORATION.  All rights reserved.
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an express
# license agreement from NVIDIA CORPORATION is strictly prohibited.

# include parent path
import numpy as np
import math

import warp as wp
from warp.tests.test_base import *

np.random.seed(42)

wp.init()


# triangulate a list of polygon face indices
def triangulate(face_counts, face_indices):
    num_tris = np.sum(np.subtract(face_counts, 2))
    num_tri_vtx = num_tris * 3
    tri_indices = np.zeros(num_tri_vtx, dtype=int)
    ctr = 0
    wedgeIdx = 0

    for nb in face_counts:
        for i in range(nb-2):
            tri_indices[ctr] = face_indices[wedgeIdx]
            tri_indices[ctr + 1] = face_indices[wedgeIdx + i + 1]
            tri_indices[ctr + 2] = face_indices[wedgeIdx + i + 2]
            ctr+=3
        wedgeIdx+=nb

    return tri_indices


@wp.kernel
def mesh_query_ray_loss(mesh: wp.uint64,
                query_points: wp.array(dtype=wp.vec3),
                query_dirs: wp.array(dtype=wp.vec3),
                intersection_points: wp.array(dtype=wp.vec3),
                loss: wp.array(dtype=float)):

    tid = wp.tid()

    p = query_points[tid]
    D = query_dirs[tid]

    max_t = 10012.0
    t = float(0.0)
    bary_u = float(0.0)
    bary_v = float(0.0)
    sign = float(0.0)
    normal = wp.vec3()
    face_index = int(0)

    q = wp.vec3()

    if wp.mesh_query_ray(mesh, p, D, max_t, t, bary_u, bary_v, sign, normal, face_index):
        q = wp.mesh_eval_position(mesh, face_index, bary_u, bary_v)

    intersection_points[tid] = q
    l = q[0]
    loss[tid] = l


def test_mesh_query_ray_grad(test, device):

    from pxr import Usd, UsdGeom, Gf, Sdf

    # test tri
    # print("Testing Single Triangle")
    # mesh_points = wp.array(np.array([[0.0, 0.0, 0.0], [2.0, 0.0, 0.0], [0.0, 2.0, 0.0]]), dtype=wp.vec3, device=device)
    # mesh_indices = wp.array(np.array([0,1,2]), dtype=int, device=device)

    mesh = Usd.Stage.Open(os.path.abspath(os.path.join(os.path.dirname(__file__), "assets/torus.usda")))
    mesh_geom = UsdGeom.Mesh(mesh.GetPrimAtPath("/World/Torus"))

    mesh_counts = mesh_geom.GetFaceVertexCountsAttr().Get()
    mesh_indices = mesh_geom.GetFaceVertexIndicesAttr().Get()

    tri_indices = triangulate(mesh_counts, mesh_indices)

    mesh_points = wp.array(np.array(mesh_geom.GetPointsAttr().Get()), dtype=wp.vec3, device=device)
    mesh_indices = wp.array(np.array(tri_indices), dtype=int, device=device)

    p = wp.vec3(50.0, 50.0, 0.0)
    D = wp.vec3(0.0, -1.0, 0.0)

    # create mesh
    mesh = wp.Mesh(
        points=mesh_points, 
        velocities=None,
        indices=mesh_indices)

    tape = wp.Tape()

    # analytic gradients
    with tape:

        query_points = wp.array(p, dtype=wp.vec3, device=device, requires_grad=True)
        query_dirs = wp.array(D, dtype=wp.vec3, device=device, requires_grad=True)
        intersection_points = wp.zeros(n=1, dtype=wp.vec3, device=device)
        loss = wp.zeros(n=1, dtype=float, device=device, requires_grad=True)

        wp.launch(kernel=mesh_query_ray_loss, dim=1, inputs=[mesh.id, query_points, query_dirs, intersection_points, loss], device=device)

    tape.backward(loss=loss)
    q = intersection_points.numpy().flatten()
    analytic_p = tape.gradients[query_points].numpy().flatten()
    analytic_D = tape.gradients[query_dirs].numpy().flatten()

    # numeric gradients

    # ray origin
    eps = 1.e-3
    loss_values_p = []
    numeric_p = np.zeros(3)

    offset_query_points = [
        wp.vec3(p[0] - eps, p[1], p[2]), wp.vec3(p[0] + eps, p[1], p[2]),
        wp.vec3(p[0], p[1] - eps, p[2]), wp.vec3(p[0], p[1] + eps, p[2]),
        wp.vec3(p[0], p[1], p[2] - eps), wp.vec3(p[0], p[1], p[2] + eps)]

    for i in range(6):
        q = offset_query_points[i]

        query_points = wp.array(q, dtype=wp.vec3, device=device)
        query_dirs = wp.array(D, dtype=wp.vec3, device=device)
        intersection_points = wp.zeros(n=1, dtype=wp.vec3, device=device)
        loss = wp.zeros(n=1, dtype=float, device=device)

        wp.launch(kernel=mesh_query_ray_loss, dim=1, inputs=[mesh.id, query_points, query_dirs, intersection_points, loss], device=device)

        loss_values_p.append(loss.numpy()[0])

    for i in range(3):
        l_0 = loss_values_p[i*2]
        l_1 = loss_values_p[i*2+1]
        gradient = (l_1 - l_0) / (2.0*eps)
        numeric_p[i] = gradient

    # ray dir
    loss_values_D = []
    numeric_D = np.zeros(3)

    offset_query_dirs = [
        wp.vec3(D[0] - eps, D[1], D[2]), wp.vec3(D[0] + eps, D[1], D[2]),
        wp.vec3(D[0], D[1] - eps, D[2]), wp.vec3(D[0], D[1] + eps, D[2]),
        wp.vec3(D[0], D[1], D[2] - eps), wp.vec3(D[0], D[1], D[2] + eps)]

    for i in range(6):
        q = offset_query_dirs[i]

        query_points = wp.array(p, dtype=wp.vec3, device=device)
        query_dirs = wp.array(q, dtype=wp.vec3, device=device)
        intersection_points = wp.zeros(n=1, dtype=wp.vec3, device=device)
        loss = wp.zeros(n=1, dtype=float, device=device)

        wp.launch(kernel=mesh_query_ray_loss, dim=1, inputs=[mesh.id, query_points, query_dirs, intersection_points, loss], device=device)

        loss_values_D.append(loss.numpy()[0])

    for i in range(3):
        l_0 = loss_values_D[i*2]
        l_1 = loss_values_D[i*2+1]
        gradient = (l_1 - l_0) / (2.0*eps)
        numeric_D[i] = gradient
    
    error_p = ((analytic_p - numeric_p) * (analytic_p - numeric_p)).sum(axis=0)
    error_D = ((analytic_D - numeric_D) * (analytic_D - numeric_D)).sum(axis=0)

    tolerance = 1.e-3
    test.assertTrue(error_p < tolerance, f"error is {error_p} which is >= {tolerance}")
    test.assertTrue(error_D < tolerance, f"error is {error_D} which is >= {tolerance}")



@wp.kernel
def raycast_kernel(mesh: wp.uint64,
                ray_starts: wp.array(dtype=wp.vec3),
                ray_directions: wp.array(dtype=wp.vec3),
                ray_hits: wp.array(dtype=wp.vec3),
                count: wp.array(dtype=int)):

    t = float(0.0)                   # hit distance along ray
    u = float(0.0)                   # hit face barycentric u
    v = float(0.0)                   # hit face barycentric v
    sign = float(0.0)                # hit face sign
    n = wp.vec3()                    # hit face normal
    f = int(0)                       # hit face index
    max_dist = 1e6                   # max raycast disance

    # ray cast against the mesh
    tid = wp.tid()

    if wp.mesh_query_ray(mesh, ray_starts[tid], ray_directions[tid], max_dist, t, u, v, sign, n, f):
        wp.atomic_add(count, 0, 1)


# tests rays against a quad of two connected triangles
# with rays exactly falling on the edge, tests that
# there are no leaks

def test_mesh_query_ray_edge(test, device):

    # Create raycast starts and directions
    xx, yy = np.meshgrid(np.arange(0.1, 0.4, 0.01), np.arange(0.1, 0.4, 0.01))
    xx = xx.flatten().reshape(-1, 1)
    yy = yy.flatten().reshape(-1, 1)
    zz = np.ones_like(xx)

    ray_starts = np.concatenate((xx, yy, zz), axis=1)
    ray_dirs = np.zeros_like(ray_starts)
    ray_dirs[:, 2] = -1.0


    # Create simple square mesh
    vertices = np.array([[0., 0., 0.],
                        [0., 0.5, 0.],
                        [0.5, 0., 0.],
                        [0.5, 0.5, 0.]], dtype=np.float32)

    triangles = np.array([[1, 0, 2],
                          [1, 2, 3]], dtype=np.int32)


    mesh = wp.Mesh(points=wp.array(vertices, dtype=wp.vec3, device=device),
                indices=wp.array(triangles.flatten(), dtype=int, device=device))
    

    counts = wp.zeros(1, dtype=int, device=device)

    n = len(ray_starts)
    
    ray_starts = wp.array(ray_starts, shape=(n,), dtype=wp.vec3, device=device)
    ray_dirs = wp.array(ray_dirs, shape=(n,), dtype=wp.vec3, device=device)
    ray_hits = wp.zeros((n, ), dtype=wp.vec3, device=device)
    
    wp.launch(kernel=raycast_kernel,
            dim=n,
            inputs=[mesh.id, ray_starts, ray_dirs, ray_hits, counts],
            device=device)
    wp.synchronize()

    test.assertEqual(counts.numpy()[0], n)


def register(parent):

    devices = wp.get_devices()

    class TestMeshQueryRay(parent):
        pass

    add_function_test(TestMeshQueryRay, "test_mesh_query_ray_edge", test_mesh_query_ray_edge, devices=devices)

    # USD import failures should not count as a test failure
    try:
        from pxr import Usd, UsdGeom
        have_usd = True
    except:
        have_usd = False

    if have_usd:
        add_function_test(TestMeshQueryRay, "test_mesh_query_ray_grad", test_mesh_query_ray_grad, devices=devices)

    return TestMeshQueryRay

if __name__ == '__main__':
    c = register(unittest.TestCase)
    unittest.main(verbosity=2)



# Copyright (c) 2022 NVIDIA CORPORATION.  All rights reserved.
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an express
# license agreement from NVIDIA CORPORATION is strictly prohibited.

import unittest
import os
import sys

import numpy as np
import warp as wp

# redirects and captures all stdout output (including from C-libs)
class StdOutCapture:

    def begin(self):
        
        # save original
        self.saved = sys.stdout
        self.target = os.dup(self.saved.fileno())
        
        # create temporary capture stream
        import io, tempfile
        self.tempfile = io.TextIOWrapper(
                            tempfile.TemporaryFile(buffering=0),
                            encoding="utf-8",
                            errors="replace",
                            newline="",
                            write_through=True)

        os.dup2(self.tempfile.fileno(), self.saved.fileno())
        
        sys.stdout = self.tempfile

    def end(self):

        os.dup2(self.target, self.saved.fileno())
        os.close(self.target)
        
        self.tempfile.seek(0)
        res = self.tempfile.buffer.read()
        self.tempfile.close()

        sys.stdout = self.saved

        return str(res.decode("utf-8"))


class CheckOutput:

    def __init__(self, test):
        self.test = test

    def __enter__(self):
        wp.force_load()

        self.capture = StdOutCapture()
        self.capture.begin()


    def __exit__(self, exc_type, exc_value, traceback):
        
        # ensure any stdout output is flushed
        wp.synchronize()

        s = self.capture.end()
        if (s != ""):
            print(s)
            
        # fail if kernel produces any stdout (e.g.: from wp.expect_eq() builtins)
        self.test.assertEqual(s, "")


def assert_array_equal(result, expect):

    a = result.numpy()
    b = expect.numpy()

    if ((a == b).all() == False):
        raise AssertionError(f"Unexpected result, got: {a} expected: {b}")

def assert_np_equal(result, expect, tol=0.0):

    a = result.flatten()
    b = expect.flatten()

    if tol == 0.0:

        if ((a == b).all() == False):
            raise AssertionError(f"Unexpected result, got: {a} expected: {b}")

    else:

        delta = a-b
        err = np.max(np.abs(delta))
        if err > tol:
            raise AssertionError(f"Maximum expected error exceeds tolerance got: {a}, expected: {b}, with err: {err} > {tol}")


def create_test_func(func, device, **kwargs):

    # pass args to func
    def test_func(self):
        func(self, device, **kwargs)

    return test_func


def sanitize_identifier(s):
    """ replace all non-identifier characters with '_' """

    s = str(s)
    if s.isidentifier():
        return s
    else:
        import re
        return re.sub('\W|^(?=\d)', '_', s)


def add_function_test(cls, name, func, devices=None, **kwargs):

    if devices is None:
        setattr(cls, name, create_test_func(func, None, **kwargs))
    else:
        for device in devices:
            setattr(cls, name + "_" + sanitize_identifier(device), create_test_func(func, device, **kwargs))


def add_kernel_test(cls, kernel, dim, name=None, expect=None, inputs=None, devices=None):
    
    def test_func(self):

        args = []
        if (inputs):
            args.extend(inputs)

        if (expect):
            # allocate outputs to match results
            result = wp.array(expect, dtype=int, device=device)
            output = wp.zeros_like(result)

            args.append(output)

        # force load so that we don't generate any log output during launch
        kernel.module.load(device)

        capture = StdOutCapture()
        capture.begin()

        with CheckOutput(self):
            wp.launch(kernel, dim=dim, inputs=args, device=device)
        
        s = capture.end()

        # fail if kernel produces any stdout (e.g.: from wp.expect_eq() builtins)
        self.assertEqual(s, "")

        # check output values
        if expect:
            assert_array_equal(output, result)

    if name == None:
        name = kernel.key

    # register test func with class for the given devices
    if devices is None:
        setattr(cls, name, test_func)
    else:
        for device in devices:
            setattr(cls, name + "_" + sanitize_identifier(device), test_func)



# Copyright (c) 2022 NVIDIA CORPORATION.  All rights reserved.
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an express
# license agreement from NVIDIA CORPORATION is strictly prohibited.

import numpy as np
import warp as wp

import math

import warp as wp
from warp.tests.test_base import *

import unittest


wp.init()


@wp.kernel
def inc(a: wp.array(dtype=float)):

    tid = wp.tid()
    a[tid] = a[tid] + 1.0


@wp.kernel
def arange(start: int, step: int, a: wp.array(dtype=int)):

    tid = wp.tid()
    a[tid] = start + step * tid


def test_multigpu_set_device(test, device):

    assert len(wp.get_cuda_devices()) > 1, "At least two CUDA devices are required"

    n = 32

    wp.set_device("cuda:0")
    a0 = wp.empty(n, dtype=int)
    wp.launch(arange, dim=a0.size, inputs=[0, 1, a0])

    wp.set_device("cuda:1")
    a1 = wp.empty(n, dtype=int)
    wp.launch(arange, dim=a1.size, inputs=[0, 1, a1])

    wp.synchronize()

    assert a0.device == "cuda:0"
    assert a1.device == "cuda:1"

    expected = np.arange(n, dtype=int)

    assert_np_equal(a0.numpy(), expected)
    assert_np_equal(a1.numpy(), expected)


def test_multigpu_scoped_device(test, device):

    assert len(wp.get_cuda_devices()) > 1, "At least two CUDA devices are required"

    n = 32

    with wp.ScopedDevice("cuda:0"):
        a0 = wp.empty(n, dtype=int)
        wp.launch(arange, dim=a0.size, inputs=[0, 1, a0])

    with wp.ScopedDevice("cuda:1"):
        a1 = wp.empty(n, dtype=int)
        wp.launch(arange, dim=a1.size, inputs=[0, 1, a1])

    wp.synchronize()

    assert a0.device == "cuda:0"
    assert a1.device == "cuda:1"

    expected = np.arange(n, dtype=int)

    assert_np_equal(a0.numpy(), expected)
    assert_np_equal(a1.numpy(), expected)


def test_multigpu_nesting(test, device):
    
    assert len(wp.get_cuda_devices()) > 1, "At least two CUDA devices are required"

    initial_device = wp.get_device()
    initial_cuda_device = wp.get_cuda_device()

    with wp.ScopedDevice("cuda:1"):
        assert wp.get_device() == "cuda:1"
        assert wp.get_cuda_device() == "cuda:1"

        with wp.ScopedDevice("cuda:0"):
            assert wp.get_device() == "cuda:0"
            assert wp.get_cuda_device() == "cuda:0"

            with wp.ScopedDevice("cpu"):
                assert wp.get_device() == "cpu"
                assert wp.get_cuda_device() == "cuda:0"

                wp.set_device("cuda:1")

                assert wp.get_device() == "cuda:1"
                assert wp.get_cuda_device() == "cuda:1"

            assert wp.get_device() == "cuda:0"
            assert wp.get_cuda_device() == "cuda:0"

        assert wp.get_device() == "cuda:1"
        assert wp.get_cuda_device() == "cuda:1"

    assert wp.get_device() == initial_device
    assert wp.get_cuda_device() == initial_cuda_device


def test_multigpu_pingpong(test, device):
    
    assert len(wp.get_cuda_devices()) > 1, "At least two CUDA devices are required"

    n = 32
    
    a0 = wp.zeros(n, dtype=float, device="cuda:0")
    a1 = wp.zeros(n, dtype=float, device="cuda:1")

    iters = 10

    for i in range(iters):

        wp.launch(inc, dim=a0.size, inputs=[a0], device=a0.device)
        wp.synchronize_device(a0.device)
        wp.copy(a1, a0)

        wp.launch(inc, dim=a1.size, inputs=[a1], device=a1.device)
        wp.synchronize_device(a1.device)
        wp.copy(a0, a1)

    wp.synchronize()

    expected = np.full(n, iters * 2, dtype=np.float32)

    assert_np_equal(a0.numpy(), expected)
    assert_np_equal(a1.numpy(), expected)


def test_multigpu_from_torch(test, device):

    import torch

    n = 32

    t0 = torch.arange(0, n, 1, dtype=torch.int32, device="cuda:0")
    t1 = torch.arange(0, n*2, 2, dtype=torch.int32, device="cuda:1")

    a0 = wp.from_torch(t0, dtype=wp.int32)
    a1 = wp.from_torch(t1, dtype=wp.int32)

    wp.synchronize()

    assert a0.device == "cuda:0"
    assert a1.device == "cuda:1"

    expected0 = np.arange(0, n, 1)
    expected1 = np.arange(0, n*2, 2)

    assert_np_equal(a0.numpy(), expected0)
    assert_np_equal(a1.numpy(), expected1)


def test_multigpu_to_torch(test, device):

    n = 32

    with wp.ScopedDevice("cuda:0"):
        a0 = wp.empty(n, dtype=wp.int32)
        wp.launch(arange, dim=a0.size, inputs=[0, 1, a0])

    with wp.ScopedDevice("cuda:1"):
        a1 = wp.empty(n, dtype=wp.int32)
        wp.launch(arange, dim=a1.size, inputs=[0, 2, a1])

    wp.synchronize()

    t0 = wp.to_torch(a0)
    t1 = wp.to_torch(a1)

    assert str(t0.device) == "cuda:0"
    assert str(t1.device) == "cuda:1"

    expected0 = np.arange(0, n, 1, dtype=np.int32)
    expected1 = np.arange(0, n*2, 2, dtype=np.int32)

    assert_np_equal(t0.cpu().numpy(), expected0)
    assert_np_equal(t1.cpu().numpy(), expected1)


def test_multigpu_torch_interop(test, device):

    import torch
    
    n = 32

    with torch.cuda.device(0):
        t0 = torch.arange(n, dtype=torch.float32, device="cuda")
        torch.cuda.synchronize()
        a0 = wp.from_torch(t0)
        wp.launch(inc, dim=a0.size, inputs=[a0], device="cuda")

    with torch.cuda.device(1):
        t1 = torch.arange(n, dtype=torch.float32, device="cuda")
        torch.cuda.synchronize()
        a1 = wp.from_torch(t1)
        wp.launch(inc, dim=a1.size, inputs=[a1], device="cuda")

    wp.synchronize()

    assert a0.device == "cuda:0"
    assert a1.device == "cuda:1"

    expected = np.arange(n, dtype=int) + 1

    # ensure the torch tensors were modified by warp
    assert_np_equal(t0.cpu().numpy(), expected)
    assert_np_equal(t1.cpu().numpy(), expected)


def register(parent):

    class TestMultigpu(parent):
        pass

    if wp.get_cuda_device_count() > 1:

        add_function_test(TestMultigpu, "test_multigpu_set_device", test_multigpu_set_device)
        add_function_test(TestMultigpu, "test_multigpu_scoped_device", test_multigpu_scoped_device)
        add_function_test(TestMultigpu, "test_multigpu_nesting", test_multigpu_nesting)
        add_function_test(TestMultigpu, "test_multigpu_pingpong", test_multigpu_pingpong)

        try:
            import torch
            add_function_test(TestMultigpu, "test_multigpu_from_torch", test_multigpu_from_torch)
            add_function_test(TestMultigpu, "test_multigpu_to_torch", test_multigpu_to_torch)
            add_function_test(TestMultigpu, "test_multigpu_torch_interop", test_multigpu_torch_interop)
        except:
            pass
    
    return TestMultigpu

if __name__ == '__main__':
    c = register(unittest.TestCase)
    unittest.main(verbosity=2, failfast=False)



# Copyright (c) 2022 NVIDIA CORPORATION.  All rights reserved.
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an express
# license agreement from NVIDIA CORPORATION is strictly prohibited.

import os

version = "0.5.0"

cuda_path = None        # path to local CUDA toolchain, if None at init time warp will attempt to find the SDK using CUDA_PATH env var

verify_fp = False       # verify inputs and outputs are finite after each launch
verify_cuda = False     # if true will check CUDA errors after each kernel launch / memory operation
print_launches = False  # if true will print out launch information

enable_backward = False # disable code gen of backwards pass

mode = "release"
verbose = False

host_compiler = None    # user can specify host compiler here, otherwise will attempt to find one automatically

cache_kernels = True
kernel_cache_dir = None # path to kernel cache directory, if None a default path will be used

ptx_target_arch = 70    # target architecture for PTX generation, defaults to the lowest architecture that supports all of Warp's features



import numpy as np
import open3d as o3d

import torch
import warp as wp

wp.config.mode = "debug"
wp.init()
device = "cuda"

@wp.kernel
def raycast_kernel(mesh: wp.uint64,
                   ray_starts: wp.array(dtype=wp.vec3),
                   ray_directions: wp.array(dtype=wp.vec3),
                   ray_hits: wp.array(dtype=wp.vec3),
                   count: int):

    #tid = wp.tid()

    t = float(0.0)                   # hit distance along ray
    u = float(0.0)                   # hit face barycentric u
    v = float(0.0)                   # hit face barycentric v
    sign = float(0.0)                # hit face sign
    n = wp.vec3()                    # hit face normal
    f = int(0)                       # hit face index
    max_dist = 1e6                   # max raycast disance


    # ray cast against the mesh
    for tid in range(count):
        if wp.mesh_query_ray(mesh, ray_starts[tid], ray_directions[tid], max_dist, t, u, v, sign, n, f):
            ray_hits[tid] = ray_starts[tid] + t * ray_directions[tid]


def ray_cast(mesh, ray_starts, ray_directions):
    n = len(ray_starts)
    ray_starts = wp.array(ray_starts.cpu().numpy(), shape=(n,), dtype=wp.vec3, device=device)
    ray_directions = wp.array(ray_directions.cpu().numpy(), shape=(n,), dtype=wp.vec3, device=device)
    ray_hits = wp.zeros((n, ), dtype=wp.vec3, device=device)
    wp.launch(kernel=raycast_kernel,
              dim=1,
              inputs=[mesh.id, ray_starts, ray_directions, ray_hits, n],
              device=device)
    wp.synchronize()
    return ray_hits.numpy()


# Create raycast starts and directions
xx, yy = torch.meshgrid(torch.arange(0.1, 0.4, 0.01, device=device), torch.arange(0.1, 0.4, 0.01, device=device))
xx = xx.flatten().view(-1, 1)
yy = yy.flatten().view(-1, 1)
zz = torch.ones_like(xx)
ray_starts = torch.cat((xx, yy, zz), dim=1)
ray_dirs = torch.zeros_like(ray_starts)
ray_dirs[:, 2] = -1.0

# Create simple square mesh
vertices = np.array([[0., 0., 0.],
                     [0., 0.5, 0.],
                     [0.5, 0., 0.],
                     [0.5, 0.5, 0.]], dtype=np.float32)

triangles = np.array([[1, 0, 2],
                     [1, 2, 3]], dtype=np.int32)

o3d_mesh = o3d.geometry.TriangleMesh()
o3d_mesh.vertices = o3d.utility.Vector3dVector(vertices)
o3d_mesh.triangles = o3d.utility.Vector3iVector(triangles)

mesh = wp.Mesh(points=wp.array(vertices, dtype=wp.vec3, device=device),
               indices=wp.array(triangles.flatten(), dtype=int, device=device))

ray_hits = ray_cast(mesh, ray_starts, ray_dirs)

pcd = o3d.geometry.PointCloud()
pcd.points = o3d.utility.Vector3dVector(ray_hits)
o3d.visualization.draw_geometries([o3d_mesh, pcd])



# Copyright (c) 2022 NVIDIA CORPORATION.  All rights reserved.
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an express
# license agreement from NVIDIA CORPORATION is strictly prohibited.

from __future__ import annotations

import os
import re
import sys
import imp
import ast
import math
import inspect
import typing
import weakref
import ctypes
import copy
import textwrap

import numpy as np

from typing import Tuple
from typing import List
from typing import Dict
from typing import Any
from typing import Callable

from warp.types import *
import warp.config

# map operator to function name
builtin_operators = {}

# see https://www.ics.uci.edu/~pattis/ICS-31/lectures/opexp.pdf for a 
# nice overview of python operators

builtin_operators[ast.Add] = "add"
builtin_operators[ast.Sub] = "sub"
builtin_operators[ast.Mult] = "mul"
builtin_operators[ast.Div] = "div"
builtin_operators[ast.FloorDiv] = "floordiv"
builtin_operators[ast.Pow] = "pow"
builtin_operators[ast.Mod] = "mod"
builtin_operators[ast.USub] = "neg"
builtin_operators[ast.Not] = "unot"

builtin_operators[ast.Gt] = ">"
builtin_operators[ast.Lt] = "<"
builtin_operators[ast.GtE] = ">="
builtin_operators[ast.LtE] = "<="
builtin_operators[ast.Eq] = "=="
builtin_operators[ast.NotEq] = "!="

class StructInstance:
    def __init__(self, struct: Struct):
        self.__dict__['_struct_'] = struct
        self.__dict__['_c_struct_'] = struct.ctype()

    def __setattr__(self, name, value):
        assert name in self._struct_.vars, "invalid struct member variable {}".format(name)
        if isinstance(self._struct_.vars[name].type, array):
            # wp.array
            assert isinstance(value, array)
            assert value.dtype == self._struct_.vars[name].type.dtype, "assign to struct member variable {} failed, expected type {}, got type {}".format(name, self._struct_.vars[name].type.dtype, value.dtype)
            setattr(self._c_struct_, name, value.__ctype__())
        elif issubclass(self._struct_.vars[name].type, ctypes.Array):
            # array type e.g. vec3
            setattr(self._c_struct_, name, self._struct_.vars[name].type(*value))
        else:
            # primitive type
            setattr(self._c_struct_, name, self._struct_.vars[name].type._type_(value))

        self.__dict__[name] = value

    def __repr__(self):
        lines = []
        for k, v in self.__dict__.items():
            if not k.startswith('_'):
                lines.append(' ' * 4 + f"{k}: {v.__repr__()}")
        return "StructInstance(\n" + "\n".join(lines) + "\n)\n"


class Struct:
    
    def __init__(self, cls, key, module):
        self.cls = cls
        self.module = module
        self.key = key

        self.vars = {}
        for label, type in self.cls.__annotations__.items():
            self.vars[label] = Var(label, type)

        fields = []
        for label, var in self.vars.items():
            if isinstance(var.type, array):
                fields.append((label, array_t))
            elif issubclass(var.type, ctypes.Array):
                fields.append((label, var.type))
            else:
                fields.append((label, var.type._type_))

        class StructType(ctypes.Structure):
            _fields_ = fields

        self.ctype = StructType

        if (module):
            module.register_struct(self)

    def __call__(self):
        '''
        This function returns s = StructInstance(self)
        s uses self.cls as template.
        To enable autocomplete on s, we inherit from self.cls.
        For example,

        @wp.struct
        class A:
            # annotations
            ...

        The type annotations are inherited in A(), allowing autocomplete in kernels
        '''
        # return StructInstance(self)

        class NewStructInstance(self.cls, StructInstance):
            def __init__(inst):
                StructInstance.__init__(inst, self)
        return NewStructInstance()

class Var:
    def __init__(self, label, type, requires_grad=False, constant=None):

        # convert built-in types to wp types
        if type == float:
            type = float32
        elif type == int:
            type = int32

        self.label = label
        self.type = type
        self.requires_grad = requires_grad
        self.constant = constant

    def __str__(self):
        return self.label

    def ctype(self):
        if (isinstance(self.type, array)):
            return f"array_t<{str(self.type.dtype.__name__)}>"
        if (isinstance(self.type, Struct)):
            return make_full_qualified_name(self.type.cls)
        else:
            return str(self.type.__name__)

class Block:
    
    # Represents a basic block of instructions, e.g.: list
    # of straight line instructions inside a for-loop or conditional 

    def __init__(self):
        
        # list of statements inside this block
        self.body_forward = []
        self.body_replay = []
        self.body_reverse = []

        # list of vars declared in this block
        self.vars = []
       

class Adjoint:

    # Source code transformer, this class takes a Python function and
    # generates forward and backward SSA forms of the function instructions

    def __init__(adj, func):

        adj.func = func
        
        # build AST from function object
        adj.source = inspect.getsource(func)

        # get source code lines and line number where function starts
        adj.raw_source, adj.fun_lineno = inspect.getsourcelines(func)

        # keep track of line number in function code
        adj.lineno = None
        
        # ensures that indented class methods can be parsed as kernels
        adj.source = textwrap.dedent(adj.source)
        
        # extract name of source file
        adj.filename = inspect.getsourcefile(func) or "unknown source file"

        # build AST
        adj.tree = ast.parse(adj.source)

        adj.fun_name = adj.tree.body[0].name

        # parse argument types
        adj.arg_types = typing.get_type_hints(func)
        adj.args = []

        for name, type in adj.arg_types.items():

            # skip return hint
            if name == "return":
                continue
            
            # add variable for argument
            arg = Var(name, type, False)
            adj.args.append(arg)


    # generate function ssa form and adjoint
    def build(adj, builder):

        adj.builder = builder

        adj.symbols = {}            # map from symbols to adjoint variables
        adj.variables = []          # list of local variables (in order)

        adj.cond = None             # condition variable if in branch
        adj.return_var = None       # return type for function or kernel

        # blocks
        adj.blocks = [Block()]

        # holds current indent level
        adj.prefix = ""

        # used to generate new label indices
        adj.label_count = 0

        # update symbol map for each argument
        for a in adj.args:
            adj.symbols[a.label] = a

        # recursively evaluate function body
        try:
            adj.eval(adj.tree.body[0])
        except Exception as e:
            try:
                lineno = adj.lineno+adj.fun_lineno
                line = adj.source.splitlines()[adj.lineno]
                msg = f"Error while parsing function \"{adj.fun_name}\" at {adj.filename}:{lineno}:\n{line}\n"
                ex, data, traceback = sys.exc_info()
                e = ex("".join([msg] + list(data.args))).with_traceback(traceback)
            finally:
                raise e

        for a in adj.args:
            if isinstance(a.type, Struct):
                builder.build_struct(a.type)

    # code generation methods
    def format_template(adj, template, input_vars, output_var):

        # output var is always the 0th index
        args = [output_var] + input_vars
        s = template.format(*args)

        return s

    # generates a comma separated list of args
    def format_args(adj, prefix, args):
        s = ""
        sep = ""

        for a in args:
            if type(a) == warp.context.Function:
                
                # functions don't have a var_ prefix so strip it off here
                if (prefix == "var_"):
                    s += sep + a.key
                else:
                    s += sep + prefix + a.key

            else:
                s += sep + prefix + str(a)

            sep = ", "

        return s

    def indent(adj):
        adj.prefix = adj.prefix + "\t"

    def dedent(adj):
        adj.prefix = adj.prefix[0:-1]  


    def begin_block(adj):
        b = Block()
        
        # give block a unique id
        b.label = adj.label_count
        adj.label_count += 1

        adj.blocks.append(b)
        return b

    def end_block(adj):
        return adj.blocks.pop()

    def add_var(adj, type=None, constant=None):
        index = len(adj.variables)

        # allocate new variable
        v = Var(str(index), type=type, constant=constant)

        adj.variables.append(v)
        
        adj.blocks[-1].vars.append(v)

        return v

    # append a statement to the forward pass
    def add_forward(adj, statement, replay=None, skip_replay=False):

        adj.blocks[-1].body_forward.append(adj.prefix + statement)

        if skip_replay == False:

            if (replay):
                # if custom replay specified the output it
                adj.blocks[-1].body_replay.append(adj.prefix + replay)
            else:
                # by default just replay the original statement
                adj.blocks[-1].body_replay.append(adj.prefix + statement)

    # append a statement to the reverse pass
    def add_reverse(adj, statement):

        adj.blocks[-1].body_reverse.append(adj.prefix + statement)


    def add_constant(adj, n):

        output = adj.add_var(type=type(n), constant=n)
        return output

    def add_load(adj, input):

        output = adj.add_var(input.type)

        adj.add_forward("var_{} = {};".format(output, input))
        adj.add_reverse("adj_{} += adj_{};".format(input, output))

        return output

    def add_comp(adj, op_strings, left, comps):
        output = adj.add_var(bool)

        s = "var_" + str(output) + " = " + ("(" * len(comps)) + "var_" + str(left) + " "
        for op, comp in zip(op_strings, comps):
            s += op + " var_" + str(comp) + ") "

        s = s.rstrip() + ";"

        adj.add_forward(s)

        return output

    def add_bool_op(adj, op_string, exprs):
        output = adj.add_var(bool)
        command = "var_" + str(output) + " = " + (" " + op_string + " ").join(["var_" + str(expr) for expr in exprs]) + ";"
        adj.add_forward(command)

        return output

    def add_call(adj, func, inputs, min_outputs=None):

        # if func is overloaded then perform overload resolution here
        # we validate argument types before they go to generated native code
        resolved_func = None

        for f in func.overloads:
            match = True

            # skip type checking for variadic functions
            if not f.variadic:

                # check argument counts match (todo: default arguments?)
                if len(f.input_types) != len(inputs):
                    match = False
                    continue

                # check argument types equal
                for i, a in enumerate(f.input_types.values()):
                    
                    # if arg type registered as Any, treat as 
                    # template allowing any type to match
                    if a == Any:
                        continue

                    # handle function refs as a special case
                    if a == Callable and type(inputs[i]) is warp.context.Function:
                        continue

                    # otherwise check arg type matches input variable type
                    if not types_equal(a, inputs[i].type):
                        match = False
                        break

            # check output dimensions match expectations
            if min_outputs:

                try:
                    value_type = f.value_func(inputs)
                    if len(value_type) != min_outputs:
                        match = False
                        continue
                except Exception as e:
                    
                    # value func may fail if the user has given 
                    # incorrect args, so we need to catch this
                    match = False
                    continue

            # found a match, use it
            if (match):
                resolved_func = f
                break

        if resolved_func == None:
            
            arg_types = []

            for x in inputs:
                if isinstance(x, Var):
                    # shorten Warp primitive type names
                    if x.type.__module__ == "warp.types":
                        arg_types.append(x.type.__name__)
                    else:
                        arg_types.append(x.type.__module__ + "." + x.type.__name__)
                
                if isinstance(x, warp.context.Function):
                    arg_types.append("function")

            raise Exception(f"Couldn't find function overload for '{func.key}' that matched inputs with types: [{', '.join(arg_types)}]")

        else:
            func = resolved_func


        # if it is a user-function then build it recursively
        if not func.is_builtin():
            adj.builder.build_function(func)

        # evaluate the function type based on inputs
        value_type = func.value_func(inputs)

        # handle expression (zero output), e.g.: void do_something();
        if (value_type == None):

            forward_call = func.namespace + "{}({});".format(func.key, adj.format_args("var_", inputs))
            
            if func.skip_replay:
                adj.add_forward(forward_call, replay="//" + forward_call)
            else:
                adj.add_forward(forward_call)

            if (len(inputs)):
                reverse_call = func.namespace + "{}({}, {});".format("adj_" + func.key, adj.format_args("var_", inputs), adj.format_args("adj_", inputs))
                adj.add_reverse(reverse_call)

            return None

        # handle multiple value functions
        elif (isinstance(value_type, list)):

            output = [adj.add_var(v) for v in value_type]
            forward_call = func.namespace + "{}({});".format(func.key, adj.format_args("var_", inputs+output))
            adj.add_forward(forward_call)

            if (len(inputs)):
                reverse_call = func.namespace + "{}({}, {}, {});".format(
                    "adj_" + func.key, adj.format_args("var_", inputs+output), adj.format_args("adj_", inputs), adj.format_args("adj_", output))
                adj.add_reverse(reverse_call)

            if len(output) == 1:
                return output[0]

            return output

        # handle simple function (one output)
        else:

            output = adj.add_var(func.value_func(inputs))

            forward_call = "var_{} = ".format(output) + func.namespace + "{}({});".format(func.key, adj.format_args("var_", inputs))

            if func.skip_replay:
                adj.add_forward(forward_call, replay="//" + forward_call)
            else:
                adj.add_forward(forward_call)
            
            if (len(inputs)):
                reverse_call = func.namespace + "{}({}, {}, {});".format(
                    "adj_" + func.key, adj.format_args("var_", inputs), adj.format_args("adj_", inputs), adj.format_args("adj_", [output]))
                adj.add_reverse(reverse_call)

            return output

    def add_return(adj, var):

        if (var == None):
            adj.add_forward("return;".format(var), "goto label{};".format(adj.label_count))
        else:
            adj.add_forward("return var_{};".format(var), "goto label{};".format(adj.label_count))
            adj.add_reverse("adj_" + str(var) + " += adj_ret;")

        adj.add_reverse("label{}:;".format(adj.label_count))

        adj.label_count += 1


    # define an if statement
    def begin_if(adj, cond):

        adj.add_forward("if (var_{}) {{".format(cond))
        adj.add_reverse("}")

        adj.indent()

    def end_if(adj, cond):

        adj.dedent()

        adj.add_forward("}")
        adj.add_reverse(f"if (var_{cond}) {{")

    def begin_else(adj, cond):

        adj.add_forward(f"if (!var_{cond}) {{")
        adj.add_reverse("}")

        adj.indent()

    def end_else(adj, cond):

        adj.dedent()

        adj.add_forward("}")
        adj.add_reverse(f"if (!var_{cond}) {{")


    # define a for-loop
    def begin_for(adj, iter):

        cond_block = adj.begin_block()
        adj.add_forward(f"for_start_{cond_block.label}:;")
        adj.indent()

        # evaluate cond
        adj.add_forward(f"if (iter_cmp(var_{iter}) == 0) goto for_end_{cond_block.label};")

        # evaluate iter
        val = adj.add_call(warp.context.builtin_functions["iter_next"], [iter])

        adj.begin_block()

        return val


    def end_for(adj, iter):
       
        body_block = adj.end_block()
        cond_block = adj.end_block()
                
        ####################
        # forward pass

        for i in cond_block.body_forward:
            adj.blocks[-1].body_forward.append(i)

        for i in body_block.body_forward:
            adj.blocks[-1].body_forward.append(i)

        adj.add_forward(f"goto for_start_{cond_block.label};", skip_replay=True)
        
        adj.dedent()
        adj.add_forward(f"for_end_{cond_block.label}:;", skip_replay=True)

        ####################
        # reverse pass
        
        reverse = []
        
        # reverse iterator
        reverse.append(adj.prefix + f"var_{iter} = wp::iter_reverse(var_{iter});")

        for i in cond_block.body_forward:
            reverse.append(i)

        # zero adjoints
        for i in body_block.vars:
             reverse.append(adj.prefix + f"\tadj_{i} = 0;")

        # replay
        for i in body_block.body_replay:
            reverse.append(i)

        # reverse
        for i in reversed(body_block.body_reverse):
            reverse.append(i)            

        reverse.append(adj.prefix + f"\tgoto for_start_{cond_block.label};")
        reverse.append(adj.prefix + f"for_end_{cond_block.label}:;")    

        adj.blocks[-1].body_reverse.extend(reversed(reverse))

    # define a while loop
    def begin_while(adj, cond):

        # evaulate condition in it's own block
        # so we can control replay
        cond_block = adj.begin_block()
        cond_block.body_forward.append(f"while_start_{cond_block.label}:;")

        c = adj.eval(cond)
        
        cond_block.body_forward.append(f"if ((var_{c}) == false) goto while_end_{cond_block.label};")

        # being block around loop
        adj.begin_block()        
        adj.indent()
        

    def end_while(adj):

        adj.dedent()     
        body_block = adj.end_block()
        cond_block = adj.end_block()

        ####################
        # forward pass

        for i in cond_block.body_forward:
            adj.blocks[-1].body_forward.append(i)

        for i in body_block.body_forward:
            adj.blocks[-1].body_forward.append(i)
        
        adj.blocks[-1].body_forward.append(f"goto while_start_{cond_block.label};")
        adj.blocks[-1].body_forward.append(f"while_end_{cond_block.label}:;")


        ####################
        # reverse pass
        reverse = []
        
        # cond
        for i in cond_block.body_forward:
            reverse.append(i)

        # zero adjoints of local vars
        for i in body_block.vars:
             reverse.append(f"adj_{i} = 0;")

        # replay
        for i in body_block.body_replay:
            reverse.append(i)

        # reverse
        for i in reversed(body_block.body_reverse):
            reverse.append(i)            

        reverse.append(f"goto while_start_{cond_block.label};")
        reverse.append(f"while_end_{cond_block.label}:;")

        # output
        adj.blocks[-1].body_reverse.extend(reversed(reverse))


    def eval(adj, node):

        if hasattr(node, "lineno"):
            adj.set_lineno(node.lineno-1)

        # top level entry point for a function
        if (isinstance(node, ast.FunctionDef)):

            out = None
            for f in node.body:
                out = adj.eval(f)

            if 'return' in adj.symbols and adj.symbols['return'] is not None:
                out = adj.symbols['return']
            else:
                out = None
                
            return out

        # if statement
        elif (isinstance(node, ast.If)):

            if len(node.body) == 0:
                return None

            # eval condition
            cond = adj.eval(node.test)

            # save symbol map
            symbols_prev = adj.symbols.copy()

            # eval body
            adj.begin_if(cond)

            for stmt in node.body:
                adj.eval(stmt)

            adj.end_if(cond)

            # detect existing symbols with conflicting definitions (variables assigned inside the branch)
            # and resolve with a phi (select) function
            for items in symbols_prev.items():

                sym = items[0]
                var1 = items[1]
                var2 = adj.symbols[sym]

                if var1 != var2:
                    # insert a phi function that selects var1, var2 based on cond
                    out = adj.add_call(warp.context.builtin_functions["select"], [cond, var1, var2])
                    adj.symbols[sym] = out


            symbols_prev = adj.symbols.copy()

            # evaluate 'else' statement as if (!cond)
            if (len(node.orelse) > 0):

                adj.begin_else(cond)
            
                for stmt in node.orelse:
                    adj.eval(stmt)
                
                adj.end_else(cond)

            # detect existing symbols with conflicting definitions (variables assigned inside the else)
            # and resolve with a phi (select) function
            for items in symbols_prev.items():

                sym = items[0]
                var1 = items[1]
                var2 = adj.symbols[sym]

                if var1 != var2:
                    # insert a phi function that selects var1, var2 based on cond
                    # note the reversed order of vars since we want to use !cond as our select
                    out = adj.add_call(warp.context.builtin_functions["select"], [cond, var2, var1])
                    adj.symbols[sym] = out


            return None

        elif (isinstance(node, ast.Compare)):
            # node.left, node.ops (list of ops), node.comparators (things to compare to)
            # e.g. (left ops[0] node.comparators[0]) ops[1] node.comparators[1]

            left = adj.eval(node.left)
            comps = [adj.eval(comp) for comp in node.comparators]
            op_strings = [builtin_operators[type(op)] for op in node.ops]

            out = adj.add_comp(op_strings, left, comps)

            return out

        elif (isinstance(node, ast.BoolOp)):
            # op, expr list values

            op = node.op
            if isinstance(op, ast.And):
                func = "&&"
            elif isinstance(op, ast.Or):
                func = "||"
            else:
                raise KeyError("Op {} is not supported".format(op))

            out = adj.add_bool_op(func, [adj.eval(expr) for expr in node.values])
            return out

        elif (isinstance(node, ast.Name)):
            # lookup symbol, if it has already been assigned to a variable then return the existing mapping
            if node.id in adj.symbols:
                return adj.symbols[node.id]

            # try and resolve the name using the functions globals context (used to lookup constants + functions)
            elif node.id in adj.func.__globals__:
                obj = adj.func.__globals__[node.id]
                
                if isinstance(obj, warp.constant):
                    # evaluate constant
                    out = adj.add_constant(obj.val)
                    adj.symbols[node.id] = out
                    return out

                elif isinstance(obj, warp.context.Function):
                    # pass back ref. to function (will be converted to name during function call)
                    return obj

                else:
                    raise TypeError(f"'{node.id}' is not a local variable, function, or warp.constant")
                
            else:
                raise KeyError("Referencing undefined symbol: " + str(node.id))

        elif (isinstance(node, ast.Attribute)):
            def attribute_to_str(node):
                if isinstance(node, ast.Name):
                    return node.id
                elif isinstance(node, ast.Attribute):
                    return attribute_to_str(node.value) + "." + node.attr
                else:
                    raise RuntimeError(f"Failed to parse attribute")

            def attribute_to_val(node, context):
                if isinstance(node, ast.Name):
                    return context[node.id]
                elif isinstance(node, ast.Attribute):
                    return getattr(attribute_to_val(node.value, context), node.attr)
                else:
                    raise RuntimeError(f"Failed to parse attribute")

            key = attribute_to_str(node)

            if key in adj.symbols:
                return adj.symbols[key]
            elif isinstance(node.value, ast.Name) and node.value.id in adj.symbols:
                
                struct = adj.symbols[node.value.id]
                
                try:
                    attr_name = struct.label + "." + node.attr
                    attr_type = struct.type.vars[node.attr].type
                except:
                    raise RuntimeError(f"Error, `{node.attr}` is not an attribute of '{node.value.id}' ({struct.type})")

                # create a Var that points to the struct attribute, i.e.: directly generates `struct.attr` when used
                out = Var(attr_name, attr_type)
                
                adj.symbols[key] = out
                return adj.symbols[key]
            else:

                # try and resolve to either a wp.constant
                # or a wp.func object
                obj = attribute_to_val(node, adj.func.__globals__)
                
                if isinstance(obj, warp.constant):
                    out = adj.add_constant(obj.val)
                    adj.symbols[key] = out          # if referencing a constant
                else:
                    raise TypeError(f"'{key}' is not a local variable, warp function, or warp constant")

                return out

        elif (isinstance(node, ast.Str)):

            # string constant
            return adj.add_constant(node.s)

        elif (isinstance(node, ast.Num)):

            # lookup constant, if it has already been assigned then return existing var
            key = (node.n, type(node.n))

            if (key in adj.symbols):
                return adj.symbols[key]
            else:
                out = adj.add_constant(node.n)
                adj.symbols[key] = out
                return out


        elif (isinstance(node, ast.BinOp)):
            # evaluate binary operator arguments
            left = adj.eval(node.left)
            right = adj.eval(node.right)

            name = builtin_operators[type(node.op)]
            func = warp.context.builtin_functions[name]

            out = adj.add_call(func, [left, right])
            return out

        elif (isinstance(node, ast.UnaryOp)):
            # evaluate unary op arguments
            arg = adj.eval(node.operand)

            name = builtin_operators[type(node.op)]
            func = warp.context.builtin_functions[name]

            out = adj.add_call(func, [arg])
            return out

        elif (isinstance(node, ast.While)):

            adj.begin_while(node.test)

            symbols_prev = adj.symbols.copy()

            # eval body
            for s in node.body:
                adj.eval(s)

            # detect symbols with conflicting definitions (assigned inside the for loop)
            for items in symbols_prev.items():

                sym = items[0]
                var1 = items[1]
                var2 = adj.symbols[sym]

                if var1 != var2:

                    if (warp.config.verbose):
                        print("Warning: detected mutated variable {} during a dynamic for-loop, this is a non-differentiable operation".format(sym))

                    if (var1.constant is not None):
                        raise Exception("Error mutating a constant {} inside a dynamic loop, use the following syntax: pi = float(3.141) to declare a dynamic variable".format(sym))
                    
                    # overwrite the old variable value (violates SSA)
                    adj.add_call(warp.context.builtin_functions["copy"], [var1, var2])

                    # reset the symbol to point to the original variable
                    adj.symbols[sym] = var1

            
            adj.end_while()


        elif (isinstance(node, ast.For)):

            def is_num(a):
                return isinstance(a, ast.Num) or (
                    isinstance(a, ast.UnaryOp) and
                    isinstance(a.op, ast.USub) and isinstance(a.operand, ast.Num))

            def eval_num(a):
                if isinstance(a, ast.Num):
                    return a.n
                if (isinstance(a, ast.UnaryOp) and
                    isinstance(a.op, ast.USub) and isinstance(a.operand, ast.Num)):
                    return -a.operand.n
                return None

            # try and unroll simple range() statements that use constant args
            unrolled = False

            if isinstance(node.iter, ast.Call) and node.iter.func.id == "range":
                
                is_constant = True
                for a in node.iter.args:

                    # if all range() arguments are numeric constants we will unroll
                    # note that this only handles trivial constants, it will not unroll
                    # constant compile-time expressions e.g.: range(0, 3*2)
                    if not is_num(a):
                        is_constant = False
                        break

                if (is_constant):

                    # range(end)
                    if len(node.iter.args) == 1:
                        start = 0
                        end = eval_num(node.iter.args[0])
                        step = 1

                    # range(start, end)
                    elif len(node.iter.args) == 2:
                        start = eval_num(node.iter.args[0])
                        end = eval_num(node.iter.args[1])
                        step = 1

                    # range(start, end, step)
                    elif len(node.iter.args) == 3:
                        start = eval_num(node.iter.args[0])
                        end = eval_num(node.iter.args[1])
                        step = eval_num(node.iter.args[2])

                    # test if we're above max unroll count
                    max_iters = abs(end-start)//abs(step)
                    max_unroll = adj.builder.options["max_unroll"]

                    if max_iters > max_unroll:

                        if (warp.config.verbose):
                            print(f"Warning: fixed-size loop count of {max_iters} is larger than the module 'max_unroll' limit of {max_unroll}, will generate dynamic loop.")
                    else:

                        # unroll
                        for i in range(start, end, step):

                            var_iter = adj.add_constant(i)
                            adj.symbols[node.target.id] = var_iter

                            # eval body
                            for s in node.body:
                                adj.eval(s)

                        unrolled = True
            
            
            # couldn't unroll so generate a dynamic loop
            if not unrolled:

                # evaluate the Iterable
                iter = adj.eval(node.iter)

                adj.symbols[node.target.id] = adj.begin_for(iter)

                # for loops should be side-effect free, here we store a copy
                symbols_prev = adj.symbols.copy()

                # eval body
                for s in node.body:
                    adj.eval(s)

                # detect symbols with conflicting definitions (assigned inside the for loop)
                for items in symbols_prev.items():

                    sym = items[0]
                    var1 = items[1]
                    var2 = adj.symbols[sym]

                    if var1 != var2:

                        if (warp.config.verbose):
                            print("Warning: detected mutated variable {} during a dynamic for-loop, this is a non-differentiable operation".format(sym))

                        if (var1.constant is not None):
                            raise Exception("Error mutating a constant {} inside a dynamic loop, use the following syntax: pi = float(3.141) to declare a dynamic variable".format(sym))
                        
                        # overwrite the old variable value (violates SSA)
                        adj.add_call(warp.context.builtin_functions["copy"], [var1, var2])

                        # reset the symbol to point to the original variable
                        adj.symbols[sym] = var1

                adj.end_for(iter)

        elif (isinstance(node, ast.Expr)):
            return adj.eval(node.value)

        elif (isinstance(node, ast.Call)):

            name = None
            
            # resolve path (e.g.: module.submodule.attr) expression to a list of module names
            path = adj.resolve_path(node.func)
            
            try:
                # try and look up path in function globals
                func = eval(".".join(path), adj.func.__globals__)

                if isinstance(func, warp.context.Function) == False:
                    raise RuntimeError()
                    
            except Exception as e:

                # try and lookup in builtins, this allows users to avoid 
                # using "wp." prefix, and also handles type constructors
                # e.g.: wp.vec3 which aren't explicitly a function object
                attr = path[-1]
                if attr in warp.context.builtin_functions:
                    func = warp.context.builtin_functions[attr]
                else:
                    raise RuntimeError(f"Could not find function {'.'.join(path)} as a built-in or user-defined function. Note that user functions must be annotated with a @wp.func decorator to be called from a kernel.")

            args = []

            # eval all arguments
            for arg in node.args:
                var = adj.eval(arg)
                args.append(var)

            # get expected return count, e.g.: for multi-assignment
            min_outputs = None
            if hasattr(node, "expects"):
                min_outputs = node.expects

            # add var with value type from the function
            out = adj.add_call(func, args, min_outputs)
            return out

        elif (isinstance(node, ast.Index)):
            # the ast.Index node appears in 3.7 versions
            # when performing array slices, e.g.: x = arr[i]
            # but in version 3.8 and higher it does not appear
            return adj.eval(node.value)

        elif (isinstance(node, ast.Subscript)):

            target = adj.eval(node.value)

            indices = []

            if isinstance(node.slice, ast.Tuple):
                # handles the x[i,j] case (Python 3.8.x upward)
                for arg in node.slice.elts:
                    var = adj.eval(arg)
                    indices.append(var)

            elif isinstance(node.slice, ast.Index) and isinstance(node.slice.value, ast.Tuple):
                # handles the x[i,j] case (Python 3.7.x)
                for arg in node.slice.value.elts:
                    var = adj.eval(arg)
                    indices.append(var)
            else:
                # simple expression, e.g.: x[i]
                var = adj.eval(node.slice)
                indices.append(var)

            if isinstance(target.type, array):
                
                if len(indices) == target.type.ndim:
                    # handles array loads (where each dimension has an index specified)
                    out = adj.add_call(warp.context.builtin_functions["load"], [target, *indices])
                else:
                    # handles array views (fewer indices than dimensions)
                    out = adj.add_call(warp.context.builtin_functions["view"], [target, *indices])

            else:
                # handles non-array type indexing, e.g: vec3, mat33, etc
                out = adj.add_call(warp.context.builtin_functions["index"], [target, *indices])

            return out

        elif (isinstance(node, ast.Assign)):

            # handle the case where we are assigning multiple output variables
            if (isinstance(node.targets[0], ast.Tuple)):

                # record the expected number of outputs on the node
                # we do this so we can decide which function to
                # call based on the number of expected outputs
                if isinstance(node.value, ast.Call):
                    node.value.expects = len(node.targets[0].elts)

                # evaluate values
                out = adj.eval(node.value)

                names = []
                for v in node.targets[0].elts:
                    if (isinstance(v, ast.Name)):
                        names.append(v.id)
                    else:
                        raise RuntimeError("Multiple return functions can only assign to simple variables, e.g.: x, y = func()")

                if len(names) != len(out):
                    raise RuntimeError("Multiple return functions need to receive all their output values, incorrect number of values to unpack (expected {}, got {})".format(len(out), len(names)))
                
                for name, rhs in zip(names, out):
                    if (name in adj.symbols):
                        if not types_equal(rhs.type, adj.symbols[name].type):
                            raise TypeError("Error, assigning to existing symbol {} ({}) with different type ({})".format(name, adj.symbols[name].type, rhs.type))

                    adj.symbols[name] = rhs

                return out

            # handles the case where we are assigning to an array index (e.g.: arr[i] = 2.0)
            elif (isinstance(node.targets[0], ast.Subscript)):
                
                target = adj.eval(node.targets[0].value)
                value = adj.eval(node.value)

                slice = node.targets[0].slice
                indices = []

                if isinstance(slice, ast.Tuple):
                    # handles the x[i, j] case (Python 3.8.x upward)
                    for arg in slice.elts:
                        var = adj.eval(arg)
                        indices.append(var)

                elif isinstance(slice, ast.Index) and isinstance(slice.value, ast.Tuple):
                    # handles the x[i, j] case (Python 3.7.x)
                    for arg in slice.value.elts:
                        var = adj.eval(arg)
                        indices.append(var)
                else:
                    # simple expression, e.g.: x[i]
                    var = adj.eval(slice)
                    indices.append(var)

                if (isinstance(target.type, array)):
                    adj.add_call(warp.context.builtin_functions["store"], [target, *indices, value])
                else:
                    raise RuntimeError("Can only subscript assign array types")

                return None

            elif (isinstance(node.targets[0], ast.Name)):

                # symbol name
                name = node.targets[0].id

                # evaluate rhs
                rhs = adj.eval(node.value)

                # check type matches if symbol already defined
                if (name in adj.symbols):
                
                    if (rhs.type != adj.symbols[name].type):
                        raise TypeError("Error, assigning to existing symbol {} ({}) with different type ({})".format(name, adj.symbols[name].type, rhs.type))

                # handle simple assignment case (a = b), where we generate a value copy rather than reference
                if isinstance(node.value, ast.Name):
                    out = adj.add_var(rhs.type)
                    adj.add_call(warp.context.builtin_functions["copy"], [out, rhs])
                else:
                    out = rhs

                # update symbol map (assumes lhs is a Name node)
                adj.symbols[name] = out
                return out

            elif (isinstance(node.targets[0], ast.Attribute)):
                raise RuntimeError("Error, assignment to member variables is not currently support (structs are immutable)")

            else:
                raise RuntimeError("Error, unsupported assignment statement.")

        elif (isinstance(node, ast.Return)):
            cond = adj.cond

            out = adj.eval(node.value)
            adj.symbols['return'] = out

            if out is not None:        # set return type of function
                return_var = out
                if adj.return_var is not None and adj.return_var.ctype() != return_var.ctype():
                    raise TypeError(f"Error, function returned different types, previous: {adj.return_var.ctype()}, new {return_var.ctype()}")
                adj.return_var = return_var

            adj.add_return(out)

            return out

        elif (isinstance(node, ast.AugAssign)):
            
            # convert inplace operations (+=, -=, etc) to ssa form, e.g.: c = a + b
            left = adj.eval(node.target)
            right = adj.eval(node.value)

            # lookup
            name = builtin_operators[type(node.op)]
            func = warp.context.builtin_functions[name]

            out = adj.add_call(func, [left, right])

            # update symbol map
            adj.symbols[node.target.id] = out

        elif (isinstance(node, ast.NameConstant)):
            if node.value == True:
                out = adj.add_constant(True)
            elif node.value == False:
                out = adj.add_constant(False)
            elif node.value == None:
                raise TypeError("None type unsupported")

            return out

        elif node is None:
            return None
        else:
            raise Exception("Error, ast node of type {} not supported".format(type(node)))



    # helper to evaluate expressions of the form
    # obj1.obj2.obj3.attr in the function's global scope
    def resolve_path(adj, node):

        modules = []

        while isinstance(node, ast.Attribute):
            modules.append(node.attr)
            node = node.value

        if (isinstance(node, ast.Name)):
            modules.append(node.id)

        # reverse list since ast presents it backward order
        return [*reversed(modules)]

    # annotate generated code with the original source code line
    def set_lineno(adj, lineno):
        if adj.lineno is None or adj.lineno != lineno:
            line = lineno + adj.fun_lineno
            source = adj.raw_source[lineno].strip().ljust(70)
            adj.add_forward(f'// {source}       <L {line}>')
            adj.add_reverse(f'// adj: {source}  <L {line}>')
        adj.lineno = lineno
        

#----------------
# code generation

cpu_module_header = '''
#include "../native/builtin.h"

// avoid namespacing of float type for casting to float type, this is to avoid wp::float(x), which is not valid in C++
#define float(x) cast_float(x)
#define adj_float(x, adj_x, adj_ret) adj_cast_float(x, adj_x, adj_ret)

#define int(x) cast_int(x)
#define adj_int(x, adj_x, adj_ret) adj_cast_int(x, adj_x, adj_ret)

using namespace wp;

'''

cuda_module_header = '''
#include "../native/builtin.h"

// avoid namespacing of float type for casting to float type, this is to avoid wp::float(x), which is not valid in C++
#define float(x) cast_float(x)
#define adj_float(x, adj_x, adj_ret) adj_cast_float(x, adj_x, adj_ret)

#define int(x) cast_int(x)
#define adj_int(x, adj_x, adj_ret) adj_cast_int(x, adj_x, adj_ret)


using namespace wp;

'''

struct_template = '''
struct {name}
{{
{struct_body}
}};

'''

cpu_function_template = '''
// {filename}:{lineno}
static {return_type} {name}({forward_args})
{{
{forward_body}
}}

// {filename}:{lineno}
static void adj_{name}({reverse_args})
{{
{reverse_body}
}}

'''

cuda_function_template = '''
// {filename}:{lineno}
static CUDA_CALLABLE {return_type} {name}({forward_args})
{{
{forward_body}
}}

// {filename}:{lineno}
static CUDA_CALLABLE void adj_{name}({reverse_args})
{{
{reverse_body}
}}

'''

cuda_kernel_template = '''

extern "C" __global__ void {name}_cuda_kernel_forward({forward_args})
{{
    int _idx = blockDim.x * blockIdx.x + threadIdx.x;
    if (_idx >= dim.size) 
        return;

    set_launch_bounds(dim);

{forward_body}
}}

extern "C" __global__ void {name}_cuda_kernel_backward({reverse_args})
{{
    int _idx = blockDim.x * blockIdx.x + threadIdx.x;
    if (_idx >= dim.size) 
        return;

    set_launch_bounds(dim);

{reverse_body}
}}

'''

cpu_kernel_template = '''

void {name}_cpu_kernel_forward({forward_args})
{{
{forward_body}
}}

void {name}_cpu_kernel_backward({reverse_args})
{{
{reverse_body}
}}

'''

cuda_module_template = '''

extern "C" {{

// Python entry points
WP_API void {name}_cuda_forward(void* stream, {forward_args})
{{
    {name}_cuda_kernel_forward<<<(dim.size + 256 - 1) / 256, 256, 0, (cudaStream_t)stream>>>({forward_params});
}}

WP_API void {name}_cuda_backward(void* stream, {reverse_args})
{{
    {name}_cuda_kernel_backward<<<(dim.size + 256 - 1) / 256, 256, 0, (cudaStream_t)stream>>>({reverse_params});
}}

}} // extern C

'''

cpu_module_template = '''

extern "C" {{

// Python CPU entry points
WP_API void {name}_cpu_forward({forward_args})
{{
    set_launch_bounds(dim);

    for (int i=0; i < dim.size; ++i)
    {{
        s_threadIdx = i;

        {name}_cpu_kernel_forward({forward_params});
    }}
}}

WP_API void {name}_cpu_backward({reverse_args})
{{
    set_launch_bounds(dim);

    for (int i=0; i < dim.size; ++i)
    {{
        s_threadIdx = i;

        {name}_cpu_kernel_backward({reverse_params});
    }}
}}

}} // extern C

'''

cuda_module_header_template = '''

extern "C" {{

// Python CUDA entry points
WP_API void {name}_cuda_forward(void* stream, {forward_args});

WP_API void {name}_cuda_backward(void* stream, {reverse_args});

}} // extern C
'''

cpu_module_header_template = '''

extern "C" {{

// Python CPU entry points
WP_API void {name}_cpu_forward({forward_args});

WP_API void {name}_cpu_backward({reverse_args});

}} // extern C
'''

# converts a constant Python value to equivalent C-repr
def constant_str(value):
    
    if type(value) == bool:
        if value:
            return "true"
        else:
            return "false"

    elif type(value) == str:
        # ensure constant strings are correctly escaped
        return "\"" + str(value.encode("unicode-escape").decode()) + "\""

    elif isinstance(value, ctypes.Array):
        return "{" + ", ".join(map(str, value)) + "}"

    else:
        # otherwise just convert constant to string
        return str(value)

def indent(args, stops=1):
    sep = ",\n"
    for i in range(stops):
        sep += "\t"

    #return sep + args.replace(", ", "," + sep)
    return sep.join(args)

# generates a C function name based on the python function name
def make_full_qualified_name(func):
    return re.sub('[^0-9a-zA-Z_]+', '', func.__qualname__.replace('.', '__'))

def codegen_struct(struct, indent=4):
    body = []
    indent_block = " " * indent
    for label, var in struct.vars.items():
        assert not (isinstance(var.type, Struct))

        body.append(var.ctype() + " " + label + ";\n")

    return struct_template.format(
        name=make_full_qualified_name(struct.cls),
        struct_body="".join([indent_block + l for l in body])
    )


def codegen_func_forward_body(adj, device='cpu', indent=4):
    body = []
    indent_block = " " * indent

    for f in adj.blocks[0].body_forward:
        body += [f + "\n"]

    return "".join([indent_block + l for l in body])


def codegen_func_forward(adj, func_type='kernel', device='cpu'):
    s = ""

    # primal vars
    s += "    //---------\n"
    s += "    // primal vars\n"

    for var in adj.variables:    
        if var.constant == None:
            s += "    " + var.ctype() + " var_" + str(var.label) + ";\n"
        else:
            s += "    const " + var.ctype() + " var_" + str(var.label) + " = " + constant_str(var.constant) + ";\n"


    # forward pass
    s += "    //---------\n"
    s += "    // forward\n"

    if device == 'cpu':
        s += codegen_func_forward_body(adj, device=device, indent=4)

    elif device == 'cuda':
        if func_type == 'kernel':
            s += codegen_func_forward_body(adj, device=device, indent=8)
        else:
            s += codegen_func_forward_body(adj, device=device, indent=4)

    return s


def codegen_func_reverse_body(adj, device='cpu', indent=4):
    body = []
    indent_block = " " * indent

    # forward pass
    body += ["//---------\n"]
    body += ["// forward\n"]

    for f in adj.blocks[0].body_replay:
        body += [f + "\n"]

    # reverse pass
    body += ["//---------\n"]
    body += ["// reverse\n"]

    for l in reversed(adj.blocks[0].body_reverse):
        body += [l + "\n"]

    body += ["return;\n"]

    return "".join([indent_block + l for l in body])


def codegen_func_reverse(adj, func_type='kernel', device='cpu'):
    s = ""

    # primal vars
    s += "    //---------\n"
    s += "    // primal vars\n"

    for var in adj.variables:
        if var.constant == None:
            s += "    " + var.ctype() + " var_" + str(var.label) + ";\n"
        else:
            s += "    const " + var.ctype() + " var_" + str(var.label) + " = " + constant_str(var.constant) + ";\n"

    # dual vars
    s += "    //---------\n"
    s += "    // dual vars\n"

    for var in adj.variables:
        s += "    " + var.ctype() + " adj_" + str(var.label) + " = 0;\n"

    if device == 'cpu':
        s += codegen_func_reverse_body(adj, device=device, indent=4)
    elif device == 'cuda':
        if func_type == 'kernel':
            s += codegen_func_reverse_body(adj, device=device, indent=8)
        else:
            s += codegen_func_reverse_body(adj, device=device, indent=4)
    else:
        raise ValueError("Device {} not supported for codegen".format(device))

    return s


def codegen_func(adj, device='cpu'):

    # forward header
    # return_type = "void"

    return_type = 'void' if adj.return_var is None else adj.return_var.ctype()

    forward_args = []
    reverse_args = []

    # forward args
    for arg in adj.args:
        forward_args.append(arg.ctype() + " var_" + arg.label)
        reverse_args.append(arg.ctype() + " var_" + arg.label)

    # reverse args
    for arg in adj.args:
        reverse_args.append(arg.ctype() + " & adj_" + arg.label)
    
    if return_type != 'void':
        reverse_args.append(return_type + " & adj_ret")

    # codegen body
    forward_body = codegen_func_forward(adj, func_type='function', device=device)
    reverse_body = codegen_func_reverse(adj, func_type='function', device=device)

    if device == 'cpu':
        template = cpu_function_template
    elif device == 'cuda':
        template = cuda_function_template
    else:
        raise ValueError("Device {} is not supported".format(device))

    s = template.format(name=make_full_qualified_name(adj.func),
                        return_type=return_type,
                        forward_args=indent(forward_args),
                        reverse_args=indent(reverse_args),
                        forward_body=forward_body,
                        reverse_body=reverse_body,
                        filename=adj.filename,
                        lineno=adj.fun_lineno)

    return s


def codegen_kernel(kernel, device, options):

    adj = kernel.adj

    forward_args = ["launch_bounds_t dim"]
    reverse_args = ["launch_bounds_t dim"]

    # forward args
    for arg in adj.args:
        forward_args.append(arg.ctype() + " var_" + arg.label)
        reverse_args.append(arg.ctype() + " var_" + arg.label)

    # reverse args
    for arg in adj.args:
        reverse_args.append(arg.ctype() + " adj_" + arg.label)

    # codegen body
    forward_body = codegen_func_forward(adj, func_type='kernel', device=device)

    if options["enable_backward"]:
        reverse_body = codegen_func_reverse(adj, func_type='kernel', device=device)
    else:
        reverse_body = ""


    if device == 'cpu':
        template = cpu_kernel_template
    elif device == 'cuda':
        template = cuda_kernel_template
    else:
        raise ValueError("Device {} is not supported".format(device))


    s = template.format(name=kernel.key,
                        forward_args=indent(forward_args),
                        reverse_args=indent(reverse_args),
                        forward_body=forward_body,
                        reverse_body=reverse_body)

    return s


def codegen_module(kernel, device='cpu'):

    adj = kernel.adj

    # build forward signature
    forward_args = ["launch_bounds_t dim"]
    forward_params = ["dim"]

    for arg in adj.args:
        forward_args.append(arg.ctype() + " var_" + arg.label)
        forward_params.append("var_" + arg.label)

    # build reverse signature
    reverse_args = [*forward_args]
    reverse_params = [*forward_params]

    for arg in adj.args:
        reverse_args.append(arg.ctype() + " adj_" + arg.label)
        reverse_params.append("adj_" + arg.label)

    if device == 'cpu':
        template = cpu_module_template
    elif device == 'cuda':
        template = cuda_module_template
    else:
        raise ValueError("Device {} is not supported".format(device))

    s = template.format(name=kernel.key,
                        forward_args=indent(forward_args),
                        reverse_args=indent(reverse_args),
                        forward_params=indent(forward_params, 3),
                        reverse_params=indent(reverse_params, 3))
    return s






# Copyright (c) 2022 NVIDIA CORPORATION.  All rights reserved.
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an express
# license agreement from NVIDIA CORPORATION is strictly prohibited.

import numpy as np
import warp as wp
from warp.tests.test_base import *

wp.init()

@wp.func
def mlp_activation(z: float):
    return wp.tanh(z)


@wp.kernel
def mlp_kernel(weights: wp.array2d(dtype=float),
               bias: wp.array(dtype=float),
               x: wp.array2d(dtype=float),
               y: wp.array2d(dtype=float)):


    wp.mlp(weights, bias, mlp_activation, wp.tid(), x, y)
    

@wp.kernel
def loss_kernel(x: wp.array2d(dtype=float),
                loss: wp.array(dtype=float)):

    i, j = wp.tid()
        
    wp.atomic_add(loss, 0, x[i,j]*x[i,j])


def test_mlp(test, device):

    np.random.seed(0)

    m = 10
    n = 200

    batches = 20000
    
    weights = wp.array(np.random.rand(m, n)*0.5 - 0.5, dtype=float, device=device)
    bias = wp.array(np.random.rand(m)*0.5 - 0.5, dtype=float, device=device)

    x = wp.array(np.random.rand(n, batches), dtype=float, device=device)
    y = wp.zeros(shape=(m,batches), device=device)

    with wp.ScopedTimer("warp", active=False):
        wp.launch(mlp_kernel, dim=batches, inputs=[weights, bias, x, y], device=device)
        wp.synchronize()

    # A*x + b
    with wp.ScopedTimer("numpy", active=False):
        expect = np.tanh(weights.numpy().reshape(m,n)@x.numpy().reshape(-1, batches) + bias.numpy().reshape(m, 1))

    result = y.numpy().reshape(-1, batches)

    assert_np_equal(result, expect, tol=1.e-6)


def create_mlp(m, n):

    import torch
    torch.manual_seed(0)

    class FeedForward(torch.nn.Module):

        def __init__(self, input_size, hidden_size):
            super(FeedForward, self).__init__()

            self.input_size = input_size
            self.hidden_size  = hidden_size
            self.fc1 = torch.nn.Linear(self.input_size, self.hidden_size)
            self.act = torch.nn.Tanh()
        
        def forward(self, x):
            out = self.fc1(x)
            out = self.act(out)
            return out

    return FeedForward(m, n)


def create_golden():

    import torch

    input_size = 32
    hidden_size = 16
    batch_size = 64

    network = create_mlp(input_size, hidden_size)

    x = torch.Tensor(np.random.rand(batch_size, input_size))
    x.requires_grad = True

    y = network.forward(x)
    y.retain_grad()

    loss = torch.inner(y.flatten(), y.flatten())
    loss.backward(retain_graph=True)

    results = {}
    results["weights"] = network.fc1.weight.cpu().detach().numpy()
    results["weights_grad"] = network.fc1.weight.grad.cpu().detach().numpy()
    results["bias"] = network.fc1.bias.cpu().detach().numpy()
    results["bias_grad"] = network.fc1.bias.grad.cpu().detach().numpy()
    results["x"] = x.cpu().detach().numpy()
    results["x_grad"] = x.grad.cpu().detach().numpy()
    results["y"] = y.cpu().detach().numpy()
    results["y_grad"] = y.grad.cpu().detach().numpy()
    results["loss"] = loss.cpu().detach().numpy()

    np.save(os.path.join(os.path.dirname(__file__), "assets/mlp_golden.npy"), results, allow_pickle=True)
    
def load_golden():
    
    return np.load(os.path.join(os.path.dirname(__file__), "assets/mlp_golden.npy"), allow_pickle=True).item()


def test_mlp_grad(test, device):

    # uncomment to re-build golden files
    #create_golden()

    results = load_golden()

    torch_weights = results["weights"]
    torch_weights_grad = results["weights_grad"]
    torch_bias = results["bias"]
    torch_bias_grad = results["bias_grad"]
    torch_x = results["x"].T
    torch_x_grad = results["x_grad"].T
    torch_y = results["y"].T
    torch_y_grad = results["y_grad"].T
    torch_loss = results["loss"].T

    weights = wp.array(torch_weights, dtype=float, device=device, requires_grad=True)
    bias = wp.array(torch_bias, dtype=float, device=device, requires_grad=True)

    x = wp.array(torch_x, dtype=float, device=device, requires_grad=True)
    y = wp.array(torch_y, dtype=float, device=device, requires_grad=True)
    y.zero_()

    loss = wp.zeros(1, dtype=float, device=device, requires_grad=True)

    m = torch_weights.shape[0]
    n = torch_weights.shape[1]
    b = torch_x.shape[1]

    tape = wp.Tape()
    with tape:
        wp.launch(mlp_kernel, dim=b, inputs=[weights, bias, x, y], device=device)
        wp.launch(loss_kernel, dim=y.shape, inputs=[y, loss], device=device)

    tape.backward(loss=loss)

    # check forward result
    assert_np_equal(y.numpy().reshape(-1, b), torch_y, tol=1.e-1)
    assert_np_equal(loss.numpy(), torch_loss, tol=1.e-1)

    # check backward result
    assert_np_equal(tape.gradients[weights].numpy().reshape(m, n), torch_weights_grad, tol=1.e-1)
    assert_np_equal(tape.gradients[bias].numpy(), torch_bias_grad, tol=1.e-1)
    assert_np_equal(tape.gradients[x].numpy().reshape(n, b), torch_x_grad, tol=1.e-1)
    assert_np_equal(tape.gradients[y].numpy().reshape(m, b), torch_y_grad, tol=1.e-1)

    


def profile_mlp_torch(device):

    import torch

    m = 128
    n = 64

    steps = 20

    for i in range(steps):       

        b = 2**i

        network = create_mlp(m, n)
          
        x = torch.Tensor(np.random.rand(b, m))

        with wp.ScopedTimer("torch_forward" + str(b)):
            y = network.forward(x)
            torch.cuda.synchronize()


    for i in range(steps):       

        b = 2**i

        network = create_mlp(m, n)
          
        x = torch.Tensor(np.random.rand(b, m))
        y = network.forward(x)

        loss = torch.norm(y)

        # run once to alloc all gradients
        loss.backward(retain_graph=True)        

        with wp.ScopedTimer("torch-backward" + str(b)):
            loss.backward()
            torch.cuda.synchronize()


def profile_mlp_warp(device):

    m = 128
    n = 64

    steps = 20

    for i in range(steps):

        b = 2**i
        
        weights = wp.array(np.random.rand(m, n)*0.5 - 0.5, dtype=float, device=device)
        bias = wp.array(np.random.rand(m)*0.5 - 0.5, dtype=float, device=device)

        x = wp.array(np.random.rand(n, b), dtype=float, device=device)
        y = wp.zeros(shape=(m,b), device=device)

        with wp.ScopedTimer("warp-forward" + str(b)):
            wp.launch(mlp_kernel, dim=b, inputs=[weights, bias, x, y], device=device)
            wp.synchronize()


    for i in range(steps):

        b = 2**i

        weights = wp.array(np.random.rand(m, n)*0.5 - 0.5, dtype=float, device=device, requires_grad=True)
        bias = wp.array(np.random.rand(m)*0.5 - 0.5, dtype=float, device=device, requires_grad=True)

        x = wp.array(np.random.rand(n, b), dtype=float, device=device, requires_grad=True)
        y = wp.zeros(shape=(m,b), device=device, requires_grad=True)

        loss = wp.zeros(1, dtype=float, device=device)

        tape = wp.Tape()
        with tape:
            wp.launch(mlp_kernel, dim=b, inputs=[weights, bias, x, y], device=device)
            wp.launch(loss_kernel, dim=y.size, inputs=[y.flatten(), loss], device=device)

        # run backward once to ensure all adjoints are allocated
        tape.backward(loss)
        wp.synchronize()

        with wp.ScopedTimer("warp-backward" + str(b)):
            tape.backward(loss)
            wp.synchronize()


# profile_mlp_warp("cuda")
# profile_mlp_torch("cuda")

def register(parent):

    devices = wp.get_devices()

    class TestMLP(parent):
        pass
    
    add_function_test(TestMLP, "test_mlp", test_mlp, devices=devices)
    add_function_test(TestMLP, "test_mlp_grad", test_mlp_grad, devices=devices)
    
    return TestMLP

if __name__ == '__main__':
    c = register(unittest.TestCase)
    unittest.main(verbosity=2, failfast=False)



# Copyright (c) 2022 NVIDIA CORPORATION.  All rights reserved.
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an express
# license agreement from NVIDIA CORPORATION is strictly prohibited.

# This file is used to test reloading module references.

import warp as wp
import warp.tests.test_reference as ref

wp.init()

@wp.kernel
def kern(expect: float):
    wp.expect_eq(ref.magic(), expect)

def run(expect, device):
    wp.launch(kern, dim=1, inputs=[expect], device=device)



# Copyright (c) 2022 NVIDIA CORPORATION.  All rights reserved.
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an express
# license agreement from NVIDIA CORPORATION is strictly prohibited.

# include parent path
import numpy as np
import math

import warp as wp
from warp.tests.test_base import *

import unittest

wp.init()


@wp.kernel
def kernel_1d(a: wp.array(dtype=int, ndim=1)):

    i = wp.tid()

    wp.expect_eq(a[i], wp.tid())

    a[i] = a[i]*2
    wp.atomic_add(a, i, 1)

    wp.expect_eq(a[i], wp.tid()*2 + 1)


def test_1d(test, device):

    dim_x = 4

    a = np.arange(0, dim_x, dtype=np.int32)

    arr = wp.array(a, device=device)

    test.assertEqual(arr.shape, a.shape)
    test.assertEqual(arr.size, a.size)
    test.assertEqual(arr.ndim, a.ndim)

    with CheckOutput(test):
        wp.launch(kernel_1d, dim=arr.size, inputs=[arr], device=device)


@wp.kernel
def kernel_2d(a: wp.array(dtype=int, ndim=2), m: int, n: int):

    i = wp.tid()//n
    j = wp.tid()%n

    wp.expect_eq(a[i,j], wp.tid())
    wp.expect_eq(a[i][j], wp.tid())

    a[i,j] = a[i,j]*2
    wp.atomic_add(a, i, j, 1)

    wp.expect_eq(a[i,j], wp.tid()*2 + 1)


def test_2d(test, device):

    dim_x = 4
    dim_y = 2

    a = np.arange(0, dim_x*dim_y, dtype=np.int32)
    a = a.reshape(dim_x, dim_y)

    arr = wp.array(a, device=device)

    test.assertEqual(arr.shape, a.shape)
    test.assertEqual(arr.size, a.size)
    test.assertEqual(arr.ndim, a.ndim)

    with CheckOutput(test):
        wp.launch(kernel_2d, dim=arr.size, inputs=[arr, dim_x, dim_y], device=device)


@wp.kernel
def kernel_3d(a: wp.array(dtype=int, ndim=3), m: int, n: int, o: int):

    i = wp.tid()//(n*o)
    j = wp.tid()%(n*o)//o
    k = wp.tid()%o

    wp.expect_eq(a[i,j,k], wp.tid())
    wp.expect_eq(a[i][j][k], wp.tid())

    a[i,j,k] = a[i,j,k]*2
    a[i][j][k] = a[i][j][k]*2
    wp.atomic_add(a, i, j, k, 1)

    wp.expect_eq(a[i,j,k], wp.tid()*4 + 1)


def test_3d(test, device):
    
    dim_x = 8
    dim_y = 4
    dim_z = 2

    a = np.arange(0, dim_x*dim_y*dim_z, dtype=np.int32)
    a = a.reshape(dim_x, dim_y, dim_z)

    arr = wp.array(a, device=device)

    test.assertEqual(arr.shape, a.shape)
    test.assertEqual(arr.size, a.size)
    test.assertEqual(arr.ndim, a.ndim)

    with CheckOutput(test):
        wp.launch(kernel_3d, dim=arr.size, inputs=[arr, dim_x, dim_y, dim_z], device=device)


@wp.kernel
def kernel_4d(a: wp.array(dtype=int, ndim=4), m: int, n: int, o: int, p: int):

    i = wp.tid()//(n*o*p)
    j = wp.tid()%(n*o*p)//(o*p)
    k = wp.tid()%(o*p)/p
    l = wp.tid()%p



    wp.expect_eq(a[i,j,k,l], wp.tid())
    wp.expect_eq(a[i][j][k][l], wp.tid())

def test_4d(test, device):
    
    dim_x = 16
    dim_y = 8
    dim_z = 4
    dim_w = 2

    a = np.arange(0, dim_x*dim_y*dim_z*dim_w, dtype=np.int32)
    a = a.reshape(dim_x, dim_y, dim_z, dim_w)

    arr = wp.array(a, device=device)

    test.assertEqual(arr.shape, a.shape)
    test.assertEqual(arr.size, a.size)
    test.assertEqual(arr.ndim, a.ndim)

    with CheckOutput(test):
        wp.launch(kernel_4d, dim=arr.size, inputs=[arr, dim_x, dim_y, dim_z, dim_w], device=device)
        

@wp.kernel
def kernel_4d_transposed(a: wp.array(dtype=int, ndim=4), m: int, n: int, o: int, p: int):

    i = wp.tid()//(n*o*p)
    j = wp.tid()%(n*o*p)//(o*p)
    k = wp.tid()%(o*p)/p
    l = wp.tid()%p

    wp.expect_eq(a[l,k,j,i], wp.tid())
    wp.expect_eq(a[l][k][j][i], wp.tid())


def test_4d_transposed(test, device):
    
    dim_x = 16
    dim_y = 8
    dim_z = 4
    dim_w = 2

    a = np.arange(0, dim_x*dim_y*dim_z*dim_w, dtype=np.int32)
    a = a.reshape(dim_x, dim_y, dim_z, dim_w)
    
    arr = wp.array(a, device=device)

    # Transpose the array manually, as using the wp.array() constructor with arr.T would make it contiguous first
    a_T = a.T
    arr_T = wp.array(
        dtype=arr.dtype, shape=a_T.shape, strides=a_T.__array_interface__["strides"],
        capacity=arr.capacity, ptr=arr.ptr, owner=False, requires_grad=arr.requires_grad, device=device)

    test.assertFalse(arr_T.is_contiguous)
    test.assertEqual(arr_T.shape, a_T.shape)
    test.assertEqual(arr_T.strides, a_T.__array_interface__["strides"])
    test.assertEqual(arr_T.size, a_T.size)
    test.assertEqual(arr_T.ndim, a_T.ndim)

    with CheckOutput(test):
        wp.launch(kernel_4d_transposed, dim=arr_T.size, inputs=[arr_T, dim_x, dim_y, dim_z, dim_w], device=device)


@wp.kernel
def lower_bound_kernel(values: wp.array(dtype=float), arr: wp.array(dtype=float), indices: wp.array(dtype=int)):
    tid = wp.tid()

    indices[tid] = wp.lower_bound(arr, values[tid])


def test_lower_bound(test, device):
    arr = wp.array(np.array([0.0, 1.0, 2.0, 3.0, 4.0, 5.0], dtype=float), dtype=float, device=device)
    values = wp.array(np.array([-0.1, 0.0, 2.5, 4.0, 5.0, 5.5], dtype=float), dtype=float, device=device)
    indices = wp.zeros(6, dtype=int, device=device)

    wp.launch(kernel=lower_bound_kernel, dim=6, inputs=[values, arr, indices], device=device)

    test.assertTrue((np.array([0, 0, 3, 4, 5, 5]) == indices.numpy()).all())


@wp.kernel
def f1(arr: wp.array(dtype=float)):
    
    wp.expect_eq(arr.shape[0], 10)


@wp.kernel
def f2(arr: wp.array2d(dtype=float)):

    wp.expect_eq(arr.shape[0], 10)
    wp.expect_eq(arr.shape[1], 20)

    slice = arr[0]
    wp.expect_eq(slice.shape[0], 20)

@wp.kernel
def f3(arr: wp.array3d(dtype=float)):

    wp.expect_eq(arr.shape[0], 10)
    wp.expect_eq(arr.shape[1], 20)
    wp.expect_eq(arr.shape[2], 30)

    slice = arr[0,0]
    wp.expect_eq(slice.shape[0], 30)

@wp.kernel
def f4(arr: wp.array4d(dtype=float)):

    wp.expect_eq(arr.shape[0], 10)
    wp.expect_eq(arr.shape[1], 20)
    wp.expect_eq(arr.shape[2], 30)
    wp.expect_eq(arr.shape[3], 40)

    slice = arr[0,0,0]
    wp.expect_eq(slice.shape[0], 40)


def test_shape(test, device):
    
    with CheckOutput(test):
        a1 = wp.zeros(dtype=float, shape=10)
        wp.launch(f1, dim=1, inputs=[a1])

        a2 = wp.zeros(dtype=float, shape=(10, 20))
        wp.launch(f2, dim=1, inputs=[a2])

        a3 = wp.zeros(dtype=float, shape=(10, 20, 30))
        wp.launch(f3, dim=1, inputs=[a3])

        a4 = wp.zeros(dtype=float, shape=(10, 20, 30, 40))
        wp.launch(f4, dim=1, inputs=[a4])



def register(parent):

    devices = wp.get_devices()

    class TestArray(parent):
        pass

    add_function_test(TestArray, "test_shape", test_shape, devices=devices)

    add_function_test(TestArray, "test_1d_array", test_1d, devices=devices)
    add_function_test(TestArray, "test_2d_array", test_2d, devices=devices)
    add_function_test(TestArray, "test_3d_array", test_3d, devices=devices)
    add_function_test(TestArray, "test_4d_array", test_4d, devices=devices)
    add_function_test(TestArray, "test_4d_array_transposed", test_4d_transposed, devices=devices)
    add_function_test(TestArray, "test_lower_bound", test_lower_bound, devices=devices)

    return TestArray

if __name__ == '__main__':
    c = register(unittest.TestCase)
    unittest.main(verbosity=2)


# Copyright (c) 2022 NVIDIA CORPORATION.  All rights reserved.
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an express
# license agreement from NVIDIA CORPORATION is strictly prohibited.

# include parent path
import numpy as np
import math

import warp as wp
from warp.tests.test_base import *

import unittest

wp.init()

@wp.func
def sqr(x: float):
    return x*x

# test nested user function calls
# and explicit return type hints
@wp.func
def cube(x: float) -> float:
    return sqr(x)*x


@wp.func
def custom(x: int):
    return x + 1

@wp.func
def custom(x: float):
    return x + 1.0

@wp.func
def custom(x: wp.vec3):
    return x + wp.vec3(1.0, 0.0, 0.0)

@wp.func
def noreturn(x: wp.vec3):
    x = x + wp.vec3(0.0, 1.0, 0.0)
    
    wp.expect_eq(x, wp.vec3(1.0, 1.0, 0.0))


@wp.kernel
def test_overload_func():
    # tests overloading a custom @wp.func


    i = custom(1)
    f = custom(1.0)
    v = custom(wp.vec3(1.0, 0.0, 0.0))

    wp.expect_eq(i, 2)
    wp.expect_eq(f, 2.0)
    wp.expect_eq(v, wp.vec3(2.0, 0.0, 0.0))

    noreturn(wp.vec3(1.0, 0.0, 0.0))



def test_func_export(test, device):
    # tests calling native functions from Python
    
    q = wp.quat_identity()
    assert_np_equal(np.array([*q]), np.array([0.0, 0.0, 0.0, 1.0]))

    r = wp.quat_from_axis_angle((1.0, 0.0, 0.0), 2.0)
    assert_np_equal(np.array([*r]), np.array([0.8414709568023682, 0.0, 0.0, 0.5403022170066833]), tol=1.e-3)

    q = wp.quat(1.0, 2.0, 3.0, 4.0)
    q = wp.normalize(q)
    assert_np_equal(np.array([*q]), np.array([0.18257418274879456, 0.3651483654975891, 0.547722578048706, 0.7302967309951782]), tol=1.e-3)

    v2 = wp.vec2(1.0, 2.0)
    v2 = wp.normalize(v2)
    assert_np_equal(np.array([*v2]), np.array([0.4472135901451111, 0.8944271802902222]), tol=1.e-3)

    v3 = wp.vec3(1.0, 2.0, 3.0)
    v3 = wp.normalize(v3)
    assert_np_equal(np.array([*v3]), np.array([0.26726123690605164, 0.5345224738121033, 0.8017836809158325]), tol=1.e-3)

    v4 = wp.vec4(1.0, 2.0, 3.0, 4.0)
    v4 = wp.normalize(v4)
    assert_np_equal(np.array([*v4]), np.array([0.18257418274879456, 0.3651483654975891, 0.547722578048706, 0.7302967309951782]), tol=1.e-3)

    t = wp.transform_identity()
    assert_np_equal(np.array([*t]), np.array([0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1.0]))

    f = wp.sin(math.pi*0.5)
    test.assertAlmostEqual(f, 1.0, places=3)



def register(parent):

    devices = wp.get_devices()

    class TestFunc(parent):
        pass

    add_kernel_test(TestFunc, kernel=test_overload_func, name="test_overload_func", dim=1, devices=devices)
    add_function_test(TestFunc, func=test_func_export, name="test_func_export", devices=["cpu"])

    return TestFunc

if __name__ == '__main__':
    c = register(unittest.TestCase)
    wp.force_load()

    unittest.main(verbosity=2)





import warp as wp
import numpy as np
from warp.tests.test_base import *

wp.init()

@wp.kernel
def intersect_tri(v0: wp.vec3,
                  v1: wp.vec3,
                  v2: wp.vec3,
                  u0: wp.vec3,
                  u1: wp.vec3,
                  u2: wp.vec3,
                  result: wp.array(dtype=int)):
    
    tid = wp.tid()

    result[0] = wp.intersect_tri_tri(v0, v1, v2, u0, u1, u2)


def test_intersect_tri(test, device):

    points_intersect = [wp.vec3(0.0, 0.0, 0.0), wp.vec3(1.0, 0.0, 0.0), wp.vec3(0.0, 0.0, 1.0),
                        wp.vec3(0.5, -0.5, 0.0), wp.vec3(0.5, -0.5, 1.0), wp.vec3(0.5, 0.5, 0.0)]


    points_separated = [wp.vec3(0.0, 0.0, 0.0), wp.vec3(1.0, 0.0, 0.0), wp.vec3(0.0, 0.0, 1.0),
                        wp.vec3(-0.5, -0.5, 0.0), wp.vec3(-0.5, -0.5, 1.0), wp.vec3(-0.5, 0.5, 0.0)]

    result = wp.zeros(1, dtype=int, device=device)

    wp.launch(intersect_tri, dim=1, inputs=[*points_intersect, result], device=device)
    assert_np_equal(result.numpy(), np.array([1]))

    wp.launch(intersect_tri, dim=1, inputs=[*points_separated, result], device=device)
    assert_np_equal(result.numpy(), np.array([0]))



def register(parent):

    devices = wp.get_devices()

    class TestIntersect(parent):
        pass
    
    add_function_test(TestIntersect, "test_intersect_tri", test_intersect_tri, devices=devices)
    
    return TestIntersect

if __name__ == '__main__':
    c = register(unittest.TestCase)
    unittest.main(verbosity=2, failfast=False)



# Copyright (c) 2022 NVIDIA CORPORATION.  All rights reserved.
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an express
# license agreement from NVIDIA CORPORATION is strictly prohibited.

import numpy as np
# import matplotlib.pyplot as plt

import warp as wp
from warp.tests.test_base import *

wp.init()

@wp.kernel
def test_kernel(
    kernel_seed: int,
    int_a: wp.array(dtype=int),
    int_ab: wp.array(dtype=int),
    float_01: wp.array(dtype=float),
    float_ab: wp.array(dtype=float)):

    tid = wp.tid()

    state = wp.rand_init(kernel_seed, tid)

    int_a[tid] = wp.randi(state)
    int_ab[tid] = wp.randi(state, 0, 100)
    float_01[tid] = wp.randf(state)
    float_ab[tid] = wp.randf(state, 0.0, 100.0)

def test_rand(test, device):

    N = 10

    int_a_device = wp.zeros(N, dtype=int, device=device)
    int_a_host = wp.zeros(N, dtype=int, device="cpu")
    int_ab_device = wp.zeros(N, dtype=int, device=device)
    int_ab_host = wp.zeros(N, dtype=int, device="cpu")

    float_01_device = wp.zeros(N, dtype=float, device=device)
    float_01_host = wp.zeros(N, dtype=float, device="cpu")
    float_ab_device = wp.zeros(N, dtype=float, device=device)
    float_ab_host = wp.zeros(N, dtype=float, device="cpu")

    seed = 42

    wp.launch(
        kernel=test_kernel,
        dim=N,
        inputs=[seed, int_a_device, int_ab_device, float_01_device, float_ab_device],
        outputs=[],
        device=device
    )

    wp.copy(int_a_host, int_a_device)
    wp.copy(int_ab_host, int_ab_device)
    wp.copy(float_01_host, float_01_device)
    wp.copy(float_ab_host, float_ab_device)
    wp.synchronize()

    int_a = int_a_host.numpy()
    int_ab = int_ab_host.numpy()
    float_01 = float_01_host.numpy()
    float_ab = float_ab_host.numpy()

    int_a_true = np.array([-575632308, 59537738, 1898992239, 442961864, -1069147335, -478445524, 1803659809, 2122909397, -1888556360, 334603718])
    int_ab_true = np.array([46, 58, 46, 83, 85, 39, 72, 99, 18, 41])
    float_01_true = np.array([0.72961855, 0.86200964, 0.28770837, 0.8187722, 0.186335, 0.6101239, 0.56432086, 0.70428324, 0.64812654, 0.27679986])
    float_ab_true = np.array([96.04259, 73.33809, 63.601555, 38.647305, 71.813896, 64.65809, 77.79791, 46.579605, 94.614456, 91.921814])

    test.assertTrue((int_a == int_a_true).all())
    test.assertTrue((int_ab == int_ab_true).all())

    err = np.max(np.abs(float_01 - float_01_true))
    test.assertTrue(err < 1e-04)

    err = np.max(np.abs(float_ab - float_ab_true))
    test.assertTrue(err < 1e-04)


@wp.kernel
def sample_cdf_kernel(kernel_seed: int, cdf: wp.array(dtype=float), samples: wp.array(dtype=int)):

    tid = wp.tid()
    state = wp.rand_init(kernel_seed, tid)

    samples[tid] = wp.sample_cdf(state, cdf)


def test_sample_cdf(test, device):

    seed = 42
    cdf = np.arange(0.0, 1.0, 0.01, dtype=float)
    cdf = cdf * cdf
    cdf = wp.array(cdf, dtype=float, device=device)
    num_samples = 1000
    samples = wp.zeros(num_samples, dtype=int, device=device)

    wp.launch(kernel=sample_cdf_kernel, dim=num_samples, inputs=[seed, cdf, samples], device=device)

    # histogram should be linear
    # plt.hist(samples.numpy())
    # plt.show()


@wp.kernel
def sampling_kernel(
    kernel_seed: int,
    triangle_samples: wp.array(dtype=wp.vec2),
    square_samples: wp.array(dtype=wp.vec2),
    ring_samples: wp.array(dtype=wp.vec2),
    disk_samples: wp.array(dtype=wp.vec2),
    sphere_surface_samples: wp.array(dtype=wp.vec3),
    sphere_samples: wp.array(dtype=wp.vec3),
    hemisphere_surface_samples: wp.array(dtype=wp.vec3),
    hemisphere_samples: wp.array(dtype=wp.vec3),
    cube_samples: wp.array(dtype=wp.vec3)):

    tid = wp.tid()
    state = wp.rand_init(kernel_seed, tid)

    triangle_samples[tid] = wp.sample_triangle(state)
    ring_samples[tid] = wp.sample_unit_ring(state)
    disk_samples[tid] = wp.sample_unit_disk(state)
    sphere_surface_samples[tid] = wp.sample_unit_sphere_surface(state)
    sphere_samples[tid] = wp.sample_unit_sphere(state)
    hemisphere_surface_samples[tid] = wp.sample_unit_hemisphere_surface(state)
    hemisphere_samples[tid] = wp.sample_unit_hemisphere(state)
    square_samples[tid] = wp.sample_unit_square(state)
    cube_samples[tid] = wp.sample_unit_cube(state)


def test_sampling_methods(test, device):
    
    seed = 42
    num_samples = 100

    triangle_samples = wp.zeros(num_samples, dtype=wp.vec2, device=device)
    square_samples = wp.zeros(num_samples, dtype=wp.vec2, device=device)
    ring_samples = wp.zeros(num_samples, dtype=wp.vec2, device=device)
    disk_samples = wp.zeros(num_samples, dtype=wp.vec2, device=device)
    sphere_surface_samples = wp.zeros(num_samples, dtype=wp.vec3, device=device)
    sphere_samples = wp.zeros(num_samples, dtype=wp.vec3, device=device)
    hemisphere_surface_samples = wp.zeros(num_samples, dtype=wp.vec3, device=device)
    hemisphere_samples = wp.zeros(num_samples, dtype=wp.vec3, device=device)
    cube_samples = wp.zeros(num_samples, dtype=wp.vec3, device=device)

    wp.launch(
        kernel=sampling_kernel,
        dim=num_samples,
        inputs=[seed, triangle_samples, square_samples, ring_samples, disk_samples, sphere_surface_samples, sphere_samples, hemisphere_surface_samples, hemisphere_samples, cube_samples],
        device=device)

    # bounds check
    test.assertTrue((triangle_samples.numpy()[:,0] <= 1.0).all())
    test.assertTrue((triangle_samples.numpy()[:,0] >= 0.0).all())
    test.assertTrue((triangle_samples.numpy()[:,1] >= 0.0).all())
    test.assertTrue((triangle_samples.numpy()[:,1] >= 0.0).all())
    test.assertTrue((square_samples.numpy()[:,0] >= -0.5).all())
    test.assertTrue((square_samples.numpy()[:,0] <= 1.5).all())
    test.assertTrue((square_samples.numpy()[:,1] >= -0.5).all())
    test.assertTrue((square_samples.numpy()[:,1] <= 0.5).all())
    test.assertTrue((cube_samples.numpy()[:,0] >= -0.5).all())
    test.assertTrue((cube_samples.numpy()[:,0] <= 0.5).all())
    test.assertTrue((cube_samples.numpy()[:,1] >= -0.5).all())
    test.assertTrue((cube_samples.numpy()[:,1] <= 0.5).all())
    test.assertTrue((cube_samples.numpy()[:,2] >= -0.5).all())
    test.assertTrue((cube_samples.numpy()[:,2] <= 0.5).all())
    test.assertTrue((hemisphere_surface_samples.numpy()[:,2] >= 0.0).all())
    test.assertTrue((hemisphere_samples.numpy()[:,2] >= 0.0).all())
    test.assertTrue((np.linalg.norm(ring_samples.numpy(), axis=1) <= 1.0 + 1e6).all())
    test.assertTrue((np.linalg.norm(disk_samples.numpy(), axis=1) <= 1.0 + 1e6).all())
    test.assertTrue((np.linalg.norm(sphere_surface_samples.numpy(), axis=1) <= 1.0 + 1e6).all())
    test.assertTrue((np.linalg.norm(sphere_samples.numpy(), axis=1) <= 1.0 + 1e6).all())
    test.assertTrue((np.linalg.norm(hemisphere_surface_samples.numpy(), axis=1) <= 1.0 + 1e6).all())
    test.assertTrue((np.linalg.norm(hemisphere_samples.numpy(), axis=1) <= 1.0 + 1e6).all())


def register(parent):

    devices = wp.get_devices()

    class TestNoise(parent):
        pass

    add_function_test(TestNoise, "test_rand", test_rand, devices=devices)
    add_function_test(TestNoise, "test_sample_cdf", test_sample_cdf, devices=devices)
    add_function_test(TestNoise, "test_sampling_methods", test_sampling_methods, devices=devices)

    return TestNoise

if __name__ == '__main__':
    c = register(unittest.TestCase)
    unittest.main(verbosity=2)


# Copyright (c) 2022 NVIDIA CORPORATION.  All rights reserved.
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an express
# license agreement from NVIDIA CORPORATION is strictly prohibited.

import numpy as np
import warp as wp
from warp.tests.test_base import *

wp.init()


@wp.kernel
def mul_constant(
    x : wp.array(dtype=float),
    y : wp.array(dtype=float)):

    tid = wp.tid()
    
    y[tid] = x[tid]*2.0

@wp.kernel
def mul_variable(
    x : wp.array(dtype=float),
    y : wp.array(dtype=float),
    z : wp.array(dtype=float)):

    tid = wp.tid()

    z[tid] = x[tid]*y[tid]

@wp.kernel
def dot_product(
    x : wp.array(dtype=float),
    y : wp.array(dtype=float),
    z : wp.array(dtype=float)):

    tid = wp.tid()

    wp.atomic_add(z, 0, x[tid]*y[tid])



def test_tape_mul_constant(test, device):
        
    dim = 8
    iters = 16
    tape = wp.Tape()

    # record onto tape
    with tape:
        
        # input data
        x0 = wp.array(np.zeros(dim), dtype=wp.float32, device=device, requires_grad=True)
        x = x0

        for i in range(iters):
        
            y = wp.empty_like(x, requires_grad=True)
            wp.launch(kernel=mul_constant, dim=dim, inputs=[x], outputs=[y], device=device)
            x = y

    # loss = wp.sum(x)
    x.grad = wp.array(np.ones(dim), device=device, dtype=wp.float32)

    # run backward
    tape.backward()

    # grad = 2.0^iters
    assert_np_equal(tape.gradients[x0].numpy(), np.ones(dim)*(2**iters))


def test_tape_mul_variable(test, device):
        
    dim = 8
    tape = wp.Tape()

    # record onto tape
    with tape:
        
        # input data
        x = wp.array(np.ones(dim)*16.0, dtype=wp.float32, device=device, requires_grad=True)
        y = wp.array(np.ones(dim)*32.0, dtype=wp.float32, device=device, requires_grad=True)
        z = wp.zeros_like(x)

        wp.launch(kernel=mul_variable, dim=dim, inputs=[x, y], outputs=[z], device=device)

    # loss = wp.sum(x)
    z.grad =wp.array(np.ones(dim), device=device, dtype=wp.float32)

    # run backward
    tape.backward()

    # grad_x=y, grad_y=x
    assert_np_equal(tape.gradients[x].numpy(), y.numpy())
    assert_np_equal(tape.gradients[y].numpy(), x.numpy())

    # run backward again with different incoming gradient
    # should accumulate the same gradients again onto output
    # so gradients = 2.0*prev
    tape.backward()

    assert_np_equal(tape.gradients[x].numpy(), y.numpy()*2.0)
    assert_np_equal(tape.gradients[y].numpy(), x.numpy()*2.0)


def test_tape_dot_product(test, device):
        
    dim = 8
    tape = wp.Tape()

    # record onto tape
    with tape:
        
        # input data
        x = wp.array(np.ones(dim)*16.0, dtype=wp.float32, device=device, requires_grad=True)
        y = wp.array(np.ones(dim)*32.0, dtype=wp.float32, device=device, requires_grad=True)
        z = wp.zeros(n=1, dtype=wp.float32, device=device, requires_grad=True)

        wp.launch(kernel=dot_product, dim=dim, inputs=[x, y], outputs=[z], device=device)

    # scalar loss
    tape.backward(loss=z)

    # grad_x=y, grad_y=x
    assert_np_equal(tape.gradients[x].numpy(), y.numpy())
    assert_np_equal(tape.gradients[y].numpy(), x.numpy())


def register(parent):

    devices = wp.get_devices()

    class TestTape(parent):
        pass

    add_function_test(TestTape, "test_tape_mul_constant", test_tape_mul_constant, devices=devices)
    add_function_test(TestTape, "test_tape_mul_variable", test_tape_mul_variable, devices=devices)
    add_function_test(TestTape, "test_tape_dot_product", test_tape_dot_product, devices=devices)

    return TestTape

if __name__ == '__main__':
    c = register(unittest.TestCase)
    unittest.main(verbosity=2)




import numpy as np
import math

import warp as wp
from warp.tests.test_base import *

import unittest

wp.init()


@wp.kernel
def eval_dense_gemm(m: int, n: int, p: int, t1: int, t2: int, A: wp.array(dtype=float), B: wp.array(dtype=float), C: wp.array(dtype=float)):
    wp.dense_gemm(m, n, p, t1, t2, A, B, C)

@wp.kernel
def eval_dense_cholesky(n: int, A: wp.array(dtype=float), regularization: float, L: wp.array(dtype=float)):
    wp.dense_chol(n, A, regularization, L)

@wp.kernel
def eval_dense_subs(n: int, L: wp.array(dtype=float), b: wp.array(dtype=float), x: wp.array(dtype=float)):
    wp.dense_subs(n, L, b, x)

# helper that propagates gradients back to A, treating L as a constant / temporary variable
# allows us to reuse the Cholesky decomposition from the forward pass
@wp.kernel
def eval_dense_solve(n: int, A: wp.array(dtype=float), L: wp.array(dtype=float), b: wp.array(dtype=float), x: wp.array(dtype=float)):
    wp.dense_solve(n, A, L, b, x)



def register(parent):

    devices = wp.get_devices()

    class TestDense(parent):
        pass

    # just testing compilation of the dense matrix routines
    # most are deprecated / WIP
    wp.force_load()

    return TestDense

if __name__ == '__main__':
    c = register(unittest.TestCase)
    unittest.main(verbosity=2)


# Copyright (c) 2022 NVIDIA CORPORATION.  All rights reserved.
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an express
# license agreement from NVIDIA CORPORATION is strictly prohibited.


from . model import State, Model, ModelBuilder, Mesh

from . model import GEO_SPHERE
from . model import GEO_BOX
from . model import GEO_CAPSULE
from . model import GEO_MESH
from . model import GEO_SDF
from . model import GEO_PLANE
from . model import GEO_NONE

from . model import JOINT_PRISMATIC
from . model import JOINT_REVOLUTE
from . model import JOINT_BALL
from . model import JOINT_FIXED
from . model import JOINT_FREE
from . model import JOINT_COMPOUND
from . model import JOINT_UNIVERSAL

from . integrator_euler import SemiImplicitIntegrator
from . integrator_euler import VariationalImplicitIntegrator

from . integrator_xpbd import XPBDIntegrator

from . collide import collide
from . articulation import eval_fk, eval_ik

from . import_mjcf import parse_mjcf
from . import_urdf import parse_urdf
from . import_snu import parse_snu


# Copyright (c) 2022 NVIDIA CORPORATION.  All rights reserved.
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an express
# license agreement from NVIDIA CORPORATION is strictly prohibited.

import numpy as np

import unittest

import warp as wp
from warp.tests.test_base import *

wp.init()
epsilon = 0.00001

@wp.kernel
def closest_point_edge_edge_kernel(
    p1: wp.array(dtype=wp.vec3),
    q1: wp.array(dtype=wp.vec3),
    p2: wp.array(dtype=wp.vec3),
    q2: wp.array(dtype=wp.vec3),
    epsilon: float,
    st0: wp.array(dtype=wp.vec3),
    c1: wp.array(dtype=wp.vec3),
    c2: wp.array(dtype=wp.vec3),
):
    tid = wp.tid()
    st = wp.closest_point_edge_edge(p1[tid], q1[tid], p2[tid], q2[tid], epsilon)
    s = st[0]
    t = st[1]
    st0[tid] = st
    c1[tid] = p1[tid] + (q1[tid] - p1[tid]) * s
    c2[tid] = p2[tid] + (q2[tid] - p2[tid]) * t


def closest_point_edge_edge_launch(p1, q1, p2, q2, epsilon, st0, c1, c2, device):
    n = len(p1)
    wp.launch(
        kernel=closest_point_edge_edge_kernel,
        dim=n,
        inputs=[p1, q1, p2, q2, epsilon],
        outputs=[st0, c1, c2],
        device=device,
    )

def run_closest_point_edge_edge(p1_h, q1_h, p2_h, q2_h, device):
    p1 = wp.array(p1_h, dtype=wp.vec3, device=device)
    q1 = wp.array(q1_h, dtype=wp.vec3, device=device)
    p2 = wp.array(p2_h, dtype=wp.vec3, device=device)
    q2 = wp.array(q2_h, dtype=wp.vec3, device=device)
    st0 = wp.empty_like(p1)
    c1 = wp.empty_like(p1)
    c2 = wp.empty_like(p1)

    closest_point_edge_edge_launch(
        p1, q1, p2, q2, epsilon, st0, c1, c2, device
    )

    wp.synchronize()
    view = st0.numpy()
    return view

def test_edge_edge_middle_crossing(test, device):
    p1_h = np.array([[0, 0, 0]])
    q1_h = np.array([[1, 1, 0]])
    p2_h = np.array([[0, 1, 0]])
    q2_h = np.array([[1, 0, 0]])

    res = run_closest_point_edge_edge(p1_h, q1_h, p2_h, q2_h, device)
    st0 = res[0]
    test.assertAlmostEqual(st0[0], 0.5)  # s value
    test.assertAlmostEqual(st0[1], 0.5)  # t value

def test_edge_edge_parallel_s1_t0(test, device):
    p1_h = np.array([[0, 0, 0]])
    q1_h = np.array([[1, 1, 0]])
    p2_h = np.array([[2, 2, 0]])
    q2_h = np.array([[3, 3, 0]])

    res = run_closest_point_edge_edge(p1_h, q1_h, p2_h, q2_h, device)
    st0 = res[0]
    test.assertAlmostEqual(st0[0], 1.0)  # s value
    test.assertAlmostEqual(st0[1], 0.0)  # t value

def test_edge_edge_parallel_s0_t1(test, device):
    p1_h = np.array([[0, 0, 0]])
    q1_h = np.array([[1, 1, 0]])
    p2_h = np.array([[-2, -2, 0]])
    q2_h = np.array([[-1, -1, 0]])

    res = run_closest_point_edge_edge(p1_h, q1_h, p2_h, q2_h, device)
    st0 = res[0]
    test.assertAlmostEqual(st0[0], 0.0)  # s value
    test.assertAlmostEqual(st0[1], 1.0)  # t value

def test_edge_edge_both_degenerate_case(test, device):
    p1_h = np.array([[0, 0, 0]])
    q1_h = np.array([[0, 0, 0]])
    p2_h = np.array([[1, 1, 1]])
    q2_h = np.array([[1, 1, 1]])

    res = run_closest_point_edge_edge(p1_h, q1_h, p2_h, q2_h, device)
    st0 = res[0]
    test.assertAlmostEqual(st0[0], 0.0)  # s value
    test.assertAlmostEqual(st0[1], 0.0)  # t value

def test_edge_edge_degenerate_first_edge(test, device):
    p1_h = np.array([[0, 0, 0]])
    q1_h = np.array([[0, 0, 0]])
    p2_h = np.array([[0, 1, 0]])
    q2_h = np.array([[1, 0, 0]])

    res = run_closest_point_edge_edge(p1_h, q1_h, p2_h, q2_h, device)
    st0 = res[0]
    test.assertAlmostEqual(st0[0], 0.0)  # s value
    test.assertAlmostEqual(st0[1], 0.5)  # t value

def test_edge_edge_degenerate_second_edge(test, device):
    p1_h = np.array([[1, 0, 0]])
    q1_h = np.array([[0, 1, 0]])
    p2_h = np.array([[1, 1, 0]])
    q2_h = np.array([[1, 1, 0]])

    res = run_closest_point_edge_edge(p1_h, q1_h, p2_h, q2_h, device)
    st0 = res[0]
    test.assertAlmostEqual(st0[0], 0.5)  # s value
    test.assertAlmostEqual(st0[1], 0.0)  # t value

def test_edge_edge_parallel(test, device):
    p1_h = np.array([[0, 0, 0]])
    q1_h = np.array([[1, 0, 0]])
    p2_h = np.array([[-0.5, 1, 0]])
    q2_h = np.array([[0.5, 1, 0]])

    res = run_closest_point_edge_edge(p1_h, q1_h, p2_h, q2_h, device)
    st0 = res[0]
    test.assertAlmostEqual(st0[0], 0.0)  # s value
    test.assertAlmostEqual(st0[1], 0.5)  # t value

def test_edge_edge_perpendicular_s1_t0(test, device):
    p1_h = np.array([[0, 0, 0]])
    q1_h = np.array([[1, 1, 0]])
    p2_h = np.array([[10, 1, 0]])
    q2_h = np.array([[11, 0, 0]])

    res = run_closest_point_edge_edge(p1_h, q1_h, p2_h, q2_h, device)
    st0 = res[0]
    test.assertAlmostEqual(st0[0], 1.0)  # s value
    test.assertAlmostEqual(st0[1], 0.0)  # t value

def test_edge_edge_perpendicular_s0_t1(test, device):
    p1_h = np.array([[0, 0, 0]])
    q1_h = np.array([[1, 1, 0]])
    p2_h = np.array([[-11, -1, 0]])
    q2_h = np.array([[-5, 0, 0]])

    res = run_closest_point_edge_edge(p1_h, q1_h, p2_h, q2_h, device)
    st0 = res[0]
    test.assertAlmostEqual(st0[0], 0.0)  # s value
    test.assertAlmostEqual(st0[1], 1.0)  # t value    


def register(parent):
        
    devices = wp.get_devices()

    class TestClosestPointEdgeEdgeMethods(parent):
        pass

    add_function_test(TestClosestPointEdgeEdgeMethods, "test_edge_edge_middle_crossing", test_edge_edge_middle_crossing, devices=devices)
    add_function_test(TestClosestPointEdgeEdgeMethods, "test_edge_edge_parallel_s1_t0", test_edge_edge_parallel_s1_t0, devices=devices)
    add_function_test(TestClosestPointEdgeEdgeMethods, "test_edge_edge_parallel_s0_t1", test_edge_edge_parallel_s0_t1, devices=devices)
    add_function_test(TestClosestPointEdgeEdgeMethods, "test_edge_edge_both_degenerate_case", test_edge_edge_both_degenerate_case, devices=devices)
    add_function_test(TestClosestPointEdgeEdgeMethods, "test_edge_edge_degenerate_first_edge", test_edge_edge_degenerate_first_edge, devices=devices)
    add_function_test(TestClosestPointEdgeEdgeMethods, "test_edge_edge_degenerate_second_edge", test_edge_edge_degenerate_second_edge, devices=devices)
    add_function_test(TestClosestPointEdgeEdgeMethods, "test_edge_edge_parallel", test_edge_edge_parallel, devices=devices)
    add_function_test(TestClosestPointEdgeEdgeMethods, "test_edge_edge_perpendicular_s1_t0", test_edge_edge_perpendicular_s1_t0, devices=devices)
    add_function_test(TestClosestPointEdgeEdgeMethods, "test_edge_edge_perpendicular_s0_t1", test_edge_edge_perpendicular_s0_t1, devices=devices)

    return TestClosestPointEdgeEdgeMethods

if __name__ == '__main__':    
    c = register(unittest.TestCase)
    unittest.main(verbosity=2)